"""Showcases colour models volume and gamut plotting examples."""

import numpy as np

from colour.plotting import (
    plot_RGB_colourspaces_gamuts,
    plot_RGB_scatter,
    colour_style,
)
from colour.utilities import message_box

message_box("Colour Models Volume and Gamut Plots")

colour_style()

message_box(
    'Plotting "ITU-R BT.709" RGB colourspace volume in the "CIE xyY" '
    "colourspace."
)
plot_RGB_colourspaces_gamuts(
    ("ITU-R BT.709",), reference_colourspace="CIE xyY"
)

print("\n")

message_box(
    'Comparing "ITU-R BT.709" and "ACEScg" RGB colourspaces volume in the '
    '"CIE L*a*b*" colourspace.'
)
plot_RGB_colourspaces_gamuts(
    ("ITU-R BT.709", "ACEScg"),
    reference_colourspace="CIE Lab",
    face_colours=(None, (0.25, 0.25, 0.25)),
    edge_colours=(None, (0.25, 0.25, 0.25)),
    edge_alpha=(1.0, 0.1),
    face_alpha=(1.0, 0.0),
)

print("\n")

message_box(
    'Plotting "ACEScg" colourspaces values in the "CIE L*a*b*" colourspace.'
)

RGB = np.random.random((32, 32, 3))

plot_RGB_scatter(
    RGB,
    "ACEScg",
    reference_colourspace="CIE Lab",
    colourspaces=("ACEScg", "ITU-R BT.709"),
    face_colours=((0.25, 0.25, 0.25), None),
    edge_colours=((0.25, 0.25, 0.25), None),
    edge_alpha=(0.1, 0.5),
    face_alpha=(0.1, 0.5),
    grid_face_colours=(0.1, 0.1, 0.1),
    grid_edge_colours=(0.1, 0.1, 0.1),
    grid_edge_alpha=0.5,
    grid_face_alpha=0.1,
)



"""
Defines the unit tests for the :mod:`colour.biochemistry.michaelis_menten`
module.
"""

import numpy as np
import unittest
from itertools import permutations

from colour.biochemistry import (
    reaction_rate_MichaelisMenten_Michaelis1913,
    substrate_concentration_MichaelisMenten_Michaelis1913,
    reaction_rate_MichaelisMenten_Abebe2017,
    substrate_concentration_MichaelisMenten_Abebe2017,
)
from colour.utilities import ignore_numpy_errors

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "TestReactionRateMichaelisMentenMichaelis1913",
    "TestSubstrateConcentrationMichaelisMentenMichaelis1913",
    "TestReactionRateMichaelisMentenAbebe2017",
    "TestSubstrateConcentrationMichaelisMentenAbebe2017",
]


class TestReactionRateMichaelisMentenMichaelis1913(unittest.TestCase):
    """
    Define :func:`colour.biochemistry.michaelis_menten.\
reaction_rate_MichaelisMenten_Michaelis1913` definition unit tests methods.
    """

    def test_reaction_rate_MichaelisMenten_Michaelis1913(self):
        """
        Test :func:`colour.biochemistry.michaelis_menten.\
reaction_rate_MichaelisMenten_Michaelis1913` definition.
        """

        self.assertAlmostEqual(
            reaction_rate_MichaelisMenten_Michaelis1913(0.25, 0.5, 0.25),
            0.250000000000000,
            places=7,
        )

        self.assertAlmostEqual(
            reaction_rate_MichaelisMenten_Michaelis1913(0.5, 0.5, 0.25),
            0.333333333333333,
            places=7,
        )

        self.assertAlmostEqual(
            reaction_rate_MichaelisMenten_Michaelis1913(0.65, 0.75, 0.35),
            0.487500000000000,
            places=7,
        )

    def test_n_dimensional_reaction_rate_MichaelisMenten_Michaelis1913(self):
        """
        Test :func:`colour.biochemistry.michaelis_menten.\
reaction_rate_MichaelisMenten_Michaelis1913` definition n-dimensional arrays
        support.
        """

        v = 0.5
        V_max = 0.5
        K_m = 0.25
        S = reaction_rate_MichaelisMenten_Michaelis1913(v, V_max, K_m)

        v = np.tile(v, (6, 1))
        S = np.tile(S, (6, 1))
        np.testing.assert_almost_equal(
            reaction_rate_MichaelisMenten_Michaelis1913(v, V_max, K_m),
            S,
            decimal=7,
        )

        V_max = np.tile(V_max, (6, 1))
        K_m = np.tile(K_m, (6, 1))
        np.testing.assert_almost_equal(
            reaction_rate_MichaelisMenten_Michaelis1913(v, V_max, K_m),
            S,
            decimal=7,
        )

        v = np.reshape(v, (2, 3, 1))
        V_max = np.reshape(V_max, (2, 3, 1))
        K_m = np.reshape(K_m, (2, 3, 1))
        S = np.reshape(S, (2, 3, 1))
        np.testing.assert_almost_equal(
            reaction_rate_MichaelisMenten_Michaelis1913(v, V_max, K_m),
            S,
            decimal=7,
        )

    @ignore_numpy_errors
    def test_nan_reaction_rate_MichaelisMenten_Michaelis1913(self):
        """
        Test :func:`colour.biochemistry.michaelis_menten.\
reaction_rate_MichaelisMenten_Michaelis1913` definition nan support.
        """

        cases = [-1.0, 0.0, 1.0, -np.inf, np.inf, np.nan]
        cases = set(permutations(cases * 3, r=3))
        for case in cases:
            v = np.array(case)
            V_max = np.array(case)
            K_m = np.array(case)
            reaction_rate_MichaelisMenten_Michaelis1913(v, V_max, K_m)


class TestSubstrateConcentrationMichaelisMentenMichaelis1913(
    unittest.TestCase
):
    """
    Define :func:`colour.biochemistry.michaelis_menten.\
reaction_rate_MichaelisMenten_Michaelis1913` definition unit tests methods.
    """

    def test_substrate_concentration_MichaelisMenten_Michaelis1913(self):
        """
        Test :func:`colour.biochemistry.michaelis_menten.\
substrate_concentration_MichaelisMenten_Michaelis1913` definition.
        """

        self.assertAlmostEqual(
            substrate_concentration_MichaelisMenten_Michaelis1913(
                0.25, 0.5, 0.25
            ),
            0.250000000000000,
            places=7,
        )

        self.assertAlmostEqual(
            substrate_concentration_MichaelisMenten_Michaelis1913(
                1 / 3, 0.5, 0.25
            ),
            0.500000000000000,
            places=7,
        )

        self.assertAlmostEqual(
            substrate_concentration_MichaelisMenten_Michaelis1913(
                0.4875, 0.75, 0.35
            ),
            0.650000000000000,
            places=7,
        )

    def test_n_dimensional_substrate_concentration_MichaelisMenten_Michaelis1913(  # noqa
        self,
    ):
        """
        Test :func:`colour.biochemistry.michaelis_menten.\
substrate_concentration_MichaelisMenten_Michaelis1913` definition n-dimensional
        arrays support.
        """

        S = 1 / 3
        V_max = 0.5
        K_m = 0.25
        v = substrate_concentration_MichaelisMenten_Michaelis1913(
            S, V_max, K_m
        )

        S = np.tile(S, (6, 1))
        v = np.tile(v, (6, 1))
        np.testing.assert_almost_equal(
            substrate_concentration_MichaelisMenten_Michaelis1913(
                S, V_max, K_m
            ),
            v,
            decimal=7,
        )

        V_max = np.tile(V_max, (6, 1))
        K_m = np.tile(K_m, (6, 1))
        np.testing.assert_almost_equal(
            substrate_concentration_MichaelisMenten_Michaelis1913(
                S, V_max, K_m
            ),
            v,
            decimal=7,
        )

        S = np.reshape(S, (2, 3, 1))
        V_max = np.reshape(V_max, (2, 3, 1))
        K_m = np.reshape(K_m, (2, 3, 1))
        v = np.reshape(v, (2, 3, 1))
        np.testing.assert_almost_equal(
            substrate_concentration_MichaelisMenten_Michaelis1913(
                S, V_max, K_m
            ),
            v,
            decimal=7,
        )

    @ignore_numpy_errors
    def test_nan_substrate_concentration_MichaelisMenten_Michaelis1913(self):
        """
        Test :func:`colour.biochemistry.michaelis_menten.\
substrate_concentration_MichaelisMenten_Michaelis1913` definition nan support.
        """

        cases = [-1.0, 0.0, 1.0, -np.inf, np.inf, np.nan]
        cases = set(permutations(cases * 3, r=3))
        for case in cases:
            s = np.array(case)
            V_max = np.array(case)
            K_m = np.array(case)
            substrate_concentration_MichaelisMenten_Michaelis1913(
                s, V_max, K_m
            )


class TestReactionRateMichaelisMentenAbebe2017(unittest.TestCase):
    """
    Define :func:`colour.biochemistry.michaelis_menten.\
reaction_rate_MichaelisMenten_Abebe2017` definition unit tests methods.
    """

    def test_reaction_rate_MichaelisMenten_Abebe2017(self):
        """
        Test :func:`colour.biochemistry.michaelis_menten.\
reaction_rate_MichaelisMenten_Abebe2017` definition.
        """

        self.assertAlmostEqual(
            reaction_rate_MichaelisMenten_Abebe2017(0.25, 0.5, 0.25, 0.25),
            0.400000000000000,
            places=7,
        )

        self.assertAlmostEqual(
            reaction_rate_MichaelisMenten_Abebe2017(0.5, 0.5, 0.25, 0.25),
            0.666666666666666,
            places=7,
        )

        self.assertAlmostEqual(
            reaction_rate_MichaelisMenten_Abebe2017(0.65, 0.75, 0.35, 0.25),
            0.951219512195122,
            places=7,
        )

    def test_n_dimensional_reaction_rate_MichaelisMenten_Abebe2017(self):
        """
        Test :func:`colour.biochemistry.michaelis_menten.\
reaction_rate_MichaelisMenten_Abebe2017` definition n-dimensional arrays
        support.
        """

        v = 0.5
        V_max = 0.5
        K_m = 0.25
        b_m = 0.25
        S = reaction_rate_MichaelisMenten_Abebe2017(v, V_max, K_m, b_m)

        v = np.tile(v, (6, 1))
        S = np.tile(S, (6, 1))
        np.testing.assert_almost_equal(
            reaction_rate_MichaelisMenten_Abebe2017(v, V_max, K_m, b_m),
            S,
            decimal=7,
        )

        V_max = np.tile(V_max, (6, 1))
        K_m = np.tile(K_m, (6, 1))
        b_m = np.tile(b_m, (6, 1))
        np.testing.assert_almost_equal(
            reaction_rate_MichaelisMenten_Abebe2017(v, V_max, K_m, b_m),
            S,
            decimal=7,
        )

        v = np.reshape(v, (2, 3, 1))
        V_max = np.reshape(V_max, (2, 3, 1))
        K_m = np.reshape(K_m, (2, 3, 1))
        b_m = np.reshape(b_m, (2, 3, 1))
        S = np.reshape(S, (2, 3, 1))
        np.testing.assert_almost_equal(
            reaction_rate_MichaelisMenten_Abebe2017(v, V_max, K_m, b_m),
            S,
            decimal=7,
        )

    @ignore_numpy_errors
    def test_nan_reaction_rate_MichaelisMenten_Abebe2017(self):
        """
        Test :func:`colour.biochemistry.michaelis_menten.\
reaction_rate_MichaelisMenten_Abebe2017` definition nan support.
        """

        cases = [-1.0, 0.0, 1.0, -np.inf, np.inf, np.nan]
        cases = set(permutations(cases * 3, r=3))
        for case in cases:
            v = np.array(case)
            V_max = np.array(case)
            K_m = np.array(case)
            b_m = np.array(case)
            reaction_rate_MichaelisMenten_Abebe2017(v, V_max, K_m, b_m)


class TestSubstrateConcentrationMichaelisMentenAbebe2017(unittest.TestCase):
    """
    Define :func:`colour.biochemistry.michaelis_menten.\
reaction_rate_MichaelisMenten_Abebe2017` definition unit tests methods.
    """

    def test_substrate_concentration_MichaelisMenten_Abebe2017(self):
        """
        Test :func:`colour.biochemistry.michaelis_menten.\
substrate_concentration_MichaelisMenten_Abebe2017` definition.
        """

        self.assertAlmostEqual(
            substrate_concentration_MichaelisMenten_Abebe2017(
                0.400000000000000, 0.5, 0.25, 0.25
            ),
            0.250000000000000,
            places=7,
        )

        self.assertAlmostEqual(
            substrate_concentration_MichaelisMenten_Abebe2017(
                0.666666666666666, 0.5, 0.25, 0.25
            ),
            0.500000000000000,
            places=7,
        )

        self.assertAlmostEqual(
            substrate_concentration_MichaelisMenten_Abebe2017(
                0.951219512195122, 0.75, 0.35, 0.25
            ),
            0.650000000000000,
            places=7,
        )

    def test_n_dimensional_substrate_concentration_MichaelisMenten_Abebe2017(  # noqa
        self,
    ):
        """
        Test :func:`colour.biochemistry.michaelis_menten.\
substrate_concentration_MichaelisMenten_Abebe2017` definition n-dimensional
        arrays support.
        """

        S = 0.400000000000000
        V_max = 0.5
        K_m = 0.25
        b_m = 0.25
        v = substrate_concentration_MichaelisMenten_Abebe2017(
            S, V_max, K_m, b_m
        )

        S = np.tile(S, (6, 1))
        v = np.tile(v, (6, 1))
        np.testing.assert_almost_equal(
            substrate_concentration_MichaelisMenten_Abebe2017(
                S, V_max, K_m, b_m
            ),
            v,
            decimal=7,
        )

        V_max = np.tile(V_max, (6, 1))
        K_m = np.tile(K_m, (6, 1))
        b_m = np.tile(b_m, (6, 1))
        np.testing.assert_almost_equal(
            substrate_concentration_MichaelisMenten_Abebe2017(
                S, V_max, K_m, b_m
            ),
            v,
            decimal=7,
        )

        S = np.reshape(S, (2, 3, 1))
        V_max = np.reshape(V_max, (2, 3, 1))
        K_m = np.reshape(K_m, (2, 3, 1))
        b_m = np.reshape(b_m, (2, 3, 1))
        v = np.reshape(v, (2, 3, 1))
        np.testing.assert_almost_equal(
            substrate_concentration_MichaelisMenten_Abebe2017(
                S, V_max, K_m, b_m
            ),
            v,
            decimal=7,
        )

    @ignore_numpy_errors
    def test_nan_substrate_concentration_MichaelisMenten_Abebe2017(self):
        """
        Test :func:`colour.biochemistry.michaelis_menten.\
substrate_concentration_MichaelisMenten_Abebe2017` definition nan support.
        """

        cases = [-1.0, 0.0, 1.0, -np.inf, np.inf, np.nan]
        cases = set(permutations(cases * 3, r=3))
        for case in cases:
            s = np.array(case)
            V_max = np.array(case)
            K_m = np.array(case)
            b_m = np.array(case)
            substrate_concentration_MichaelisMenten_Abebe2017(
                s, V_max, K_m, b_m
            )


if __name__ == "__main__":
    unittest.main()



"""
Geometry Primitive Vertices
===========================

Defines various geometry primitive vertices generation methods:

-   :func:`colour.geometry.primitive_vertices_quad_mpl`
-   :func:`colour.geometry.primitive_vertices_grid_mpl`
-   :func:`colour.geometry.primitive_vertices_cube_mpl`
-   :func:`colour.geometry.primitive_vertices_sphere`
-   :func:`colour.PRIMITIVE_VERTICES_METHODS`
-   :func:`colour.primitive_vertices`
"""

from __future__ import annotations

import numpy as np

from colour.algebra import spherical_to_cartesian
from colour.geometry import MAPPING_PLANE_TO_AXIS
from colour.hints import (
    Any,
    ArrayLike,
    Boolean,
    Floating,
    Integer,
    List,
    Literal,
    NDArray,
    Optional,
    Union,
)
from colour.utilities import (
    CaseInsensitiveMapping,
    as_float_array,
    filter_kwargs,
    full,
    ones,
    tsplit,
    tstack,
    zeros,
    validate_method,
)

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "primitive_vertices_quad_mpl",
    "primitive_vertices_grid_mpl",
    "primitive_vertices_cube_mpl",
    "primitive_vertices_sphere",
    "PRIMITIVE_VERTICES_METHODS",
    "primitive_vertices",
]


def primitive_vertices_quad_mpl(
    width: Floating = 1,
    height: Floating = 1,
    depth: Floating = 0,
    origin: ArrayLike = np.array([0, 0]),
    axis: Union[Literal["+z", "+x", "+y", "yz", "xz", "xy"], str] = "+z",
) -> NDArray:
    """
    Return the vertices of a quad primitive for use with *Matplotlib*
    :class:`mpl_toolkits.mplot3d.art3d.Poly3DCollection` class.

    Parameters
    ----------
    width
        Quad width.
    height
        Quad height.
    depth
        Quad depth.
    origin
        Quad origin on the construction plane.
    axis
        Axis the quad will be normal to, or plane the quad will be co-planar
        with.

    Returns
    -------
    :class:`numpy.ndarray`
        Quad primitive vertices.

    Examples
    --------
    >>> primitive_vertices_quad_mpl()
    array([[ 0.,  0.,  0.],
           [ 1.,  0.,  0.],
           [ 1.,  1.,  0.],
           [ 0.,  1.,  0.]])
    """

    axis = MAPPING_PLANE_TO_AXIS.get(axis, axis).lower()
    axis = validate_method(
        axis, ["+x", "+y", "+z"], '"{0}" axis invalid, it must be one of {1}!'
    )

    u, v = tsplit(origin)

    if axis == "+z":
        vertices = [
            (u, v, depth),
            (u + width, v, depth),
            (u + width, v + height, depth),
            (u, v + height, depth),
        ]
    elif axis == "+y":
        vertices = [
            (u, depth, v),
            (u + width, depth, v),
            (u + width, depth, v + height),
            (u, depth, v + height),
        ]
    elif axis == "+x":
        vertices = [
            (depth, u, v),
            (depth, u + width, v),
            (depth, u + width, v + height),
            (depth, u, v + height),
        ]

    return as_float_array(vertices)


def primitive_vertices_grid_mpl(
    width: Floating = 1,
    height: Floating = 1,
    depth: Floating = 0,
    width_segments: Integer = 1,
    height_segments: Integer = 1,
    origin: ArrayLike = np.array([0, 0]),
    axis: Union[Literal["+z", "+x", "+y", "yz", "xz", "xy"], str] = "+z",
) -> NDArray:
    """
    Return the vertices of a grid primitive made of quad primitives for use
    with *Matplotlib* :class:`mpl_toolkits.mplot3d.art3d.Poly3DCollection`
    class.

    Parameters
    ----------
    width
        Grid width.
    height
        Grid height.
    depth
        Grid depth.
    width_segments:
        Grid width segments, quad primitive counts along the width.
    height_segments:
        Grid height segments, quad primitive counts along the height.
    origin
        Grid origin on the construction plane.
    axis
        Axis the grid will be normal to, or plane the grid will be co-planar
        with.

    Returns
    -------
    :class:`numpy.ndarray`
        Grid primitive vertices.

    Examples
    --------
    >>> primitive_vertices_grid_mpl(width_segments=2, height_segments=2)
    array([[[ 0. ,  0. ,  0. ],
            [ 0.5,  0. ,  0. ],
            [ 0.5,  0.5,  0. ],
            [ 0. ,  0.5,  0. ]],
    <BLANKLINE>
           [[ 0. ,  0.5,  0. ],
            [ 0.5,  0.5,  0. ],
            [ 0.5,  1. ,  0. ],
            [ 0. ,  1. ,  0. ]],
    <BLANKLINE>
           [[ 0.5,  0. ,  0. ],
            [ 1. ,  0. ,  0. ],
            [ 1. ,  0.5,  0. ],
            [ 0.5,  0.5,  0. ]],
    <BLANKLINE>
           [[ 0.5,  0.5,  0. ],
            [ 1. ,  0.5,  0. ],
            [ 1. ,  1. ,  0. ],
            [ 0.5,  1. ,  0. ]]])
    """

    u, v = tsplit(origin)

    w_x, h_y = width / width_segments, height / height_segments

    quads = []
    for i in range(width_segments):
        for j in range(height_segments):
            quads.append(
                primitive_vertices_quad_mpl(
                    w_x, h_y, depth, (i * w_x + u, j * h_y + v), axis
                )
            )

    return as_float_array(quads)


def primitive_vertices_cube_mpl(
    width: Floating = 1,
    height: Floating = 1,
    depth: Floating = 1,
    width_segments: Integer = 1,
    height_segments: Integer = 1,
    depth_segments: Integer = 1,
    origin: ArrayLike = np.array([0, 0, 0]),
    planes: Optional[
        Literal[
            "-x",
            "+x",
            "-y",
            "+y",
            "-z",
            "+z",
            "xy",
            "xz",
            "yz",
            "yx",
            "zx",
            "zy",
        ]
    ] = None,
) -> NDArray:
    """
    Return the vertices of a cube primitive made of grid primitives for use
    with *Matplotlib* :class:`mpl_toolkits.mplot3d.art3d.Poly3DCollection`
    class.

    Parameters
    ----------
    width
        Cube width.
    height
        Cube height.
    depth
        Cube depth.
    width_segments
        Cube segments count along the width.
    height_segments
        Cube segments count along the height.
    depth_segments
        Cube segments count along the depth.
    origin
        Cube origin.
    planes
        Grid primitives to include in the cube construction.

    Returns
    -------
    :class:`numpy.ndarray`
        Cube primitive vertices.

    Examples
    --------
    >>> primitive_vertices_cube_mpl()
    array([[[ 0.,  0.,  0.],
            [ 1.,  0.,  0.],
            [ 1.,  1.,  0.],
            [ 0.,  1.,  0.]],
    <BLANKLINE>
           [[ 0.,  0.,  1.],
            [ 1.,  0.,  1.],
            [ 1.,  1.,  1.],
            [ 0.,  1.,  1.]],
    <BLANKLINE>
           [[ 0.,  0.,  0.],
            [ 1.,  0.,  0.],
            [ 1.,  0.,  1.],
            [ 0.,  0.,  1.]],
    <BLANKLINE>
           [[ 0.,  1.,  0.],
            [ 1.,  1.,  0.],
            [ 1.,  1.,  1.],
            [ 0.,  1.,  1.]],
    <BLANKLINE>
           [[ 0.,  0.,  0.],
            [ 0.,  1.,  0.],
            [ 0.,  1.,  1.],
            [ 0.,  0.,  1.]],
    <BLANKLINE>
           [[ 1.,  0.,  0.],
            [ 1.,  1.,  0.],
            [ 1.,  1.,  1.],
            [ 1.,  0.,  1.]]])
    """

    axis = (
        sorted(list(MAPPING_PLANE_TO_AXIS.values()))
        if planes is None
        else [
            MAPPING_PLANE_TO_AXIS.get(plane, plane).lower() for plane in planes
        ]
    )

    u, v, w = tsplit(origin)

    w_s, h_s, d_s = width_segments, height_segments, depth_segments

    grids: List = []
    if "-z" in axis:
        grids.extend(
            primitive_vertices_grid_mpl(
                width, depth, v, w_s, d_s, (u, w), "+z"
            )
        )
    if "+z" in axis:
        grids.extend(
            primitive_vertices_grid_mpl(
                width, depth, v + height, w_s, d_s, (u, w), "+z"
            )
        )

    if "-y" in axis:
        grids.extend(
            primitive_vertices_grid_mpl(
                width, height, w, w_s, h_s, (u, v), "+y"
            )
        )
    if "+y" in axis:
        grids.extend(
            primitive_vertices_grid_mpl(
                width, height, w + depth, w_s, h_s, (u, v), "+y"
            )
        )

    if "-x" in axis:
        grids.extend(
            primitive_vertices_grid_mpl(
                depth, height, u, d_s, h_s, (w, v), "+x"
            )
        )
    if "+x" in axis:
        grids.extend(
            primitive_vertices_grid_mpl(
                depth, height, u + width, d_s, h_s, (w, v), "+x"
            )
        )

    return as_float_array(grids)


def primitive_vertices_sphere(
    radius: Floating = 0.5,
    segments: Integer = 8,
    intermediate: Boolean = False,
    origin: ArrayLike = np.array([0, 0, 0]),
    axis: Union[Literal["+z", "+x", "+y", "yz", "xz", "xy"], str] = "+z",
) -> NDArray:
    """
    Return the vertices of a latitude-longitude sphere primitive.

    Parameters
    ----------
    radius
        Sphere radius.
    segments
        Latitude-longitude segments, if the ``intermediate`` argument is
        *True*, then the sphere will have one less segment along its longitude.
    intermediate
        Whether to generate the sphere vertices at the center of the faces
        outlined by the segments of a regular sphere generated without
        the ``intermediate`` argument set to *True*. The resulting sphere is
        inscribed on the regular sphere faces but possesses the same poles.
    origin
        Sphere origin on the construction plane.
    axis
        Axis (or normal of the plane) the poles of the sphere will be aligned
        with.

    Returns
    -------
    :class:`numpy.ndarray`
        Sphere primitive vertices.

    Notes
    -----
    -   The sphere poles have latitude segments count - 1 co-located vertices.

    Examples
    --------
    >>> primitive_vertices_sphere(segments=4)  # doctest: +ELLIPSIS
    array([[[  0.0000000...e+00,   0.0000000...e+00,   5.0000000...e-01],
            [ -3.5355339...e-01,  -4.3297802...e-17,   3.5355339...e-01],
            [ -5.0000000...e-01,  -6.1232340...e-17,   3.0616170...e-17],
            [ -3.5355339...e-01,  -4.3297802...e-17,  -3.5355339...e-01],
            [ -6.1232340...e-17,  -7.4987989...e-33,  -5.0000000...e-01]],
    <BLANKLINE>
           [[  0.0000000...e+00,   0.0000000...e+00,   5.0000000...e-01],
            [  2.1648901...e-17,  -3.5355339...e-01,   3.5355339...e-01],
            [  3.0616170...e-17,  -5.0000000...e-01,   3.0616170...e-17],
            [  2.1648901...e-17,  -3.5355339...e-01,  -3.5355339...e-01],
            [  3.7493994...e-33,  -6.1232340...e-17,  -5.0000000...e-01]],
    <BLANKLINE>
           [[  0.0000000...e+00,   0.0000000...e+00,   5.0000000...e-01],
            [  3.5355339...e-01,   0.0000000...e+00,   3.5355339...e-01],
            [  5.0000000...e-01,   0.0000000...e+00,   3.0616170...e-17],
            [  3.5355339...e-01,   0.0000000...e+00,  -3.5355339...e-01],
            [  6.1232340...e-17,   0.0000000...e+00,  -5.0000000...e-01]],
    <BLANKLINE>
           [[  0.0000000...e+00,   0.0000000...e+00,   5.0000000...e-01],
            [  2.1648901...e-17,   3.5355339...e-01,   3.5355339...e-01],
            [  3.0616170...e-17,   5.0000000...e-01,   3.0616170...e-17],
            [  2.1648901...e-17,   3.5355339...e-01,  -3.5355339...e-01],
            [  3.7493994...e-33,   6.1232340...e-17,  -5.0000000...e-01]]])
    """

    axis = MAPPING_PLANE_TO_AXIS.get(axis, axis).lower()
    axis = validate_method(
        axis, ["+x", "+y", "+z"], '"{0}" axis invalid, it must be one of {1}!'
    )

    if not intermediate:
        theta = np.tile(
            np.radians(np.linspace(0, 180, segments + 1)),
            (int(segments) + 1, 1),
        )
        phi = np.transpose(
            np.tile(
                np.radians(np.linspace(-180, 180, segments + 1)),
                (int(segments) + 1, 1),
            )
        )
    else:
        theta = np.tile(
            np.radians(np.linspace(0, 180, segments * 2 + 1)[1::2][1:-1]),
            (int(segments) + 1, 1),
        )
        theta = np.hstack(
            [
                zeros((segments + 1, 1)),
                theta,
                full((segments + 1, 1), np.pi),
            ]
        )
        phi = np.transpose(
            np.tile(
                np.radians(np.linspace(-180, 180, segments + 1))
                + np.radians(360 / segments / 2),
                (int(segments), 1),
            )
        )

    rho = ones(phi.shape) * radius
    rho_theta_phi = tstack([rho, theta, phi])

    vertices = spherical_to_cartesian(rho_theta_phi)

    # Removing extra longitude vertices.
    vertices = vertices[:-1, :, :]

    if axis == "+z":
        pass
    elif axis == "+y":
        vertices = np.roll(vertices, 2, -1)
    elif axis == "+x":
        vertices = np.roll(vertices, 1, -1)

    vertices += origin

    return vertices


PRIMITIVE_VERTICES_METHODS: CaseInsensitiveMapping = CaseInsensitiveMapping(
    {
        "Quad MPL": primitive_vertices_quad_mpl,
        "Grid MPL": primitive_vertices_grid_mpl,
        "Cube MPL": primitive_vertices_cube_mpl,
        "Sphere": primitive_vertices_sphere,
    }
)
PRIMITIVE_VERTICES_METHODS.__doc__ = """
Supported geometry primitive vertices generation methods.
"""


def primitive_vertices(
    method: Union[
        Literal["Cube MPL", "Quad MPL", "Grid MPL", "Sphere"], str
    ] = "Cube MPL",
    **kwargs: Any,
) -> NDArray:
    """
    Return the vertices of a geometry primitive using given method.

    Parameters
    ----------
    method
        Vertices generation method.

    Other Parameters
    ----------------
    axis
        {:func:`colour.geometry.primitive_vertices_quad_mpl`,
        :func:`colour.geometry.primitive_vertices_grid_mpl`,
        :func:`colour.geometry.primitive_vertices_sphere`},
        **{'+z', '+x', '+y', 'yz', 'xz', 'xy'}**,
        Axis the primitive will be normal to, or plane the primitive will be
        co-planar with.
    depth
        {:func:`colour.geometry.primitive_vertices_quad_mpl`,
        :func:`colour.geometry.primitive_vertices_grid_mpl`,
        :func:`colour.geometry.primitive_vertices_cube_mpl`},
        Primitive depth.
    depth_segments
        {:func:`colour.geometry.primitive_vertices_grid_mpl`,
        :func:`colour.geometry.primitive_vertices_cube_mpl`},
        Primitive depth segments, quad primitive counts along the depth.
    height
        {:func:`colour.geometry.primitive_vertices_quad_mpl`,
        :func:`colour.geometry.primitive_vertices_grid_mpl`,
        :func:`colour.geometry.primitive_vertices_cube_mpl`},
        Primitive height.
    height_segments
        {:func:`colour.geometry.primitive_vertices_grid_mpl`,
        :func:`colour.geometry.primitive_vertices_cube_mpl`},
        Primitive height segments, quad primitive counts along the height.
    intermediate
        {:func:`colour.geometry.primitive_vertices_sphere`},
        Whether to generate the sphere vertices at the center of the faces
        outlined by the segments of a regular sphere generated without
        the ``intermediate`` argument set to *True*. The resulting sphere is
        inscribed on the regular sphere faces but possesses the same poles.
    origin
        {:func:`colour.geometry.primitive_vertices_quad_mpl`,
        :func:`colour.geometry.primitive_vertices_grid_mpl`,
        :func:`colour.geometry.primitive_vertices_cube_mpl`,
        :func:`colour.geometry.primitive_vertices_sphere`},
        Primitive origin on the construction plane.
    planes
        {:func:`colour.geometry.primitive_vertices_cube_mpl`},
        **{'-x', '+x', '-y', '+y', '-z', '+z',
        'xy', 'xz', 'yz', 'yx', 'zx', 'zy'}**,
        Included grid primitives in the cube construction.
    radius
        {:func:`colour.geometry.primitive_vertices_sphere`},
        Sphere radius.
    segments
        {:func:`colour.geometry.primitive_vertices_sphere`},
        Latitude-longitude segments, if the ``intermediate`` argument is
        *True*, then the sphere will have one less segment along its longitude.
    width
        {:func:`colour.geometry.primitive_vertices_quad_mpl`,
        :func:`colour.geometry.primitive_vertices_grid_mpl`,
        :func:`colour.geometry.primitive_vertices_cube_mpl`},
        Primitive width.
    width_segments
        {:func:`colour.geometry.primitive_vertices_grid_mpl`,
        :func:`colour.geometry.primitive_vertices_cube_mpl`},
        Primitive width segments, quad primitive counts along the width.

    Returns
    -------
    :class:`numpy.ndarray`
        Primitive vertices.

    Examples
    --------
    >>> primitive_vertices()
    array([[[ 0.,  0.,  0.],
            [ 1.,  0.,  0.],
            [ 1.,  1.,  0.],
            [ 0.,  1.,  0.]],
    <BLANKLINE>
           [[ 0.,  0.,  1.],
            [ 1.,  0.,  1.],
            [ 1.,  1.,  1.],
            [ 0.,  1.,  1.]],
    <BLANKLINE>
           [[ 0.,  0.,  0.],
            [ 1.,  0.,  0.],
            [ 1.,  0.,  1.],
            [ 0.,  0.,  1.]],
    <BLANKLINE>
           [[ 0.,  1.,  0.],
            [ 1.,  1.,  0.],
            [ 1.,  1.,  1.],
            [ 0.,  1.,  1.]],
    <BLANKLINE>
           [[ 0.,  0.,  0.],
            [ 0.,  1.,  0.],
            [ 0.,  1.,  1.],
            [ 0.,  0.,  1.]],
    <BLANKLINE>
           [[ 1.,  0.,  0.],
            [ 1.,  1.,  0.],
            [ 1.,  1.,  1.],
            [ 1.,  0.,  1.]]])
    >>> primitive_vertices('Quad MPL')
    array([[ 0.,  0.,  0.],
           [ 1.,  0.,  0.],
           [ 1.,  1.,  0.],
           [ 0.,  1.,  0.]])
    >>> primitive_vertices('Sphere', segments=4)  # doctest: +ELLIPSIS
    array([[[  0.0000000...e+00,   0.0000000...e+00,   5.0000000...e-01],
            [ -3.5355339...e-01,  -4.3297802...e-17,   3.5355339...e-01],
            [ -5.0000000...e-01,  -6.1232340...e-17,   3.0616170...e-17],
            [ -3.5355339...e-01,  -4.3297802...e-17,  -3.5355339...e-01],
            [ -6.1232340...e-17,  -7.4987989...e-33,  -5.0000000...e-01]],
    <BLANKLINE>
           [[  0.0000000...e+00,   0.0000000...e+00,   5.0000000...e-01],
            [  2.1648901...e-17,  -3.5355339...e-01,   3.5355339...e-01],
            [  3.0616170...e-17,  -5.0000000...e-01,   3.0616170...e-17],
            [  2.1648901...e-17,  -3.5355339...e-01,  -3.5355339...e-01],
            [  3.7493994...e-33,  -6.1232340...e-17,  -5.0000000...e-01]],
    <BLANKLINE>
           [[  0.0000000...e+00,   0.0000000...e+00,   5.0000000...e-01],
            [  3.5355339...e-01,   0.0000000...e+00,   3.5355339...e-01],
            [  5.0000000...e-01,   0.0000000...e+00,   3.0616170...e-17],
            [  3.5355339...e-01,   0.0000000...e+00,  -3.5355339...e-01],
            [  6.1232340...e-17,   0.0000000...e+00,  -5.0000000...e-01]],
    <BLANKLINE>
           [[  0.0000000...e+00,   0.0000000...e+00,   5.0000000...e-01],
            [  2.1648901...e-17,   3.5355339...e-01,   3.5355339...e-01],
            [  3.0616170...e-17,   5.0000000...e-01,   3.0616170...e-17],
            [  2.1648901...e-17,   3.5355339...e-01,  -3.5355339...e-01],
            [  3.7493994...e-33,   6.1232340...e-17,  -5.0000000...e-01]]])
    """

    method = validate_method(method, PRIMITIVE_VERTICES_METHODS)

    function = PRIMITIVE_VERTICES_METHODS[method]

    return function(**filter_kwargs(function, **kwargs))



"""Defines the unit tests for the :mod:`colour.algebra.extrapolation` module."""

import numpy as np
import unittest
from itertools import permutations

from colour.algebra import (
    Extrapolator,
    LinearInterpolator,
    CubicSplineInterpolator,
    PchipInterpolator,
)
from colour.utilities import ignore_numpy_errors

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "TestExtrapolator",
]


class TestExtrapolator(unittest.TestCase):
    """
    Define :class:`colour.algebra.extrapolation.Extrapolator` class unit
    tests methods.
    """

    def test_required_attributes(self):
        """Test the presence of required attributes."""

        required_attributes = ("interpolator",)

        for attribute in required_attributes:
            self.assertIn(attribute, dir(Extrapolator))

    def test_required_methods(self):
        """Test the presence of required methods."""

        required_methods = ("__init__",)

        for method in required_methods:  # pragma: no cover
            self.assertIn(method, dir(Extrapolator))

    def test_interpolator(self):
        """
        Test :attr:`colour.algebra.extrapolation.Extrapolator.interpolator`
        property.
        """

        extrapolator = Extrapolator(
            LinearInterpolator(np.array([5, 6, 7]), np.array([5, 6, 7]))
        )
        self.assertIsInstance(extrapolator.interpolator, LinearInterpolator)

    def test_method(self):
        """
        Test :attr:`colour.algebra.extrapolation.Extrapolator.method`
        property.
        """

        extrapolator = Extrapolator(
            LinearInterpolator(np.array([5, 6, 7]), np.array([5, 6, 7]))
        )
        self.assertEqual(extrapolator.method, "linear")

        extrapolator = Extrapolator(
            LinearInterpolator(np.array([5, 6, 7]), np.array([5, 6, 7])),
            method="Constant",
        )
        self.assertEqual(extrapolator.method, "constant")

    def test_left(self):
        """
        Test :attr:`colour.algebra.extrapolation.Extrapolator.left`
        property.
        """

        extrapolator = Extrapolator(
            LinearInterpolator(np.array([5, 6, 7]), np.array([5, 6, 7])),
            left=0,
        )
        self.assertEqual(extrapolator.left, 0)

    def test_right(self):
        """
        Test :attr:`colour.algebra.extrapolation.Extrapolator.right`
        property.
        """

        extrapolator = Extrapolator(
            LinearInterpolator(np.array([5, 6, 7]), np.array([5, 6, 7])),
            right=0,
        )
        self.assertEqual(extrapolator.right, 0)

    def test__call__(self):
        """
        Test :meth:`colour.algebra.extrapolation.Extrapolator.__call__`
        method.
        """

        extrapolator = Extrapolator(
            LinearInterpolator(np.array([5, 6, 7]), np.array([5, 6, 7]))
        )
        np.testing.assert_almost_equal(extrapolator((4, 8)), (4, 8))
        self.assertEqual(extrapolator(4), 4)

        extrapolator = Extrapolator(
            LinearInterpolator(np.array([3, 4, 5]), np.array([1, 2, 3])),
            method="Constant",
        )
        np.testing.assert_almost_equal(
            extrapolator((0.1, 0.2, 8, 9)), (1, 1, 3, 3)
        )
        self.assertEqual(extrapolator(0.1), 1.0)

        extrapolator = Extrapolator(
            LinearInterpolator(np.array([3, 4, 5]), np.array([1, 2, 3])),
            method="Constant",
            left=0,
        )
        np.testing.assert_almost_equal(
            extrapolator((0.1, 0.2, 8, 9)), (0, 0, 3, 3)
        )
        self.assertEqual(extrapolator(0.1), 0)

        extrapolator = Extrapolator(
            LinearInterpolator(np.array([3, 4, 5]), np.array([1, 2, 3])),
            method="Constant",
            right=0,
        )
        np.testing.assert_almost_equal(
            extrapolator((0.1, 0.2, 8, 9)), (1, 1, 0, 0)
        )
        self.assertEqual(extrapolator(9), 0)

        extrapolator = Extrapolator(
            CubicSplineInterpolator(
                np.array([3, 4, 5, 6]), np.array([1, 2, 3, 4])
            )
        )
        np.testing.assert_almost_equal(
            extrapolator((0.1, 0.2, 8.0, 9.0)), (-1.9, -1.8, 6.0, 7.0)
        )
        self.assertEqual(extrapolator(9), 7)

        extrapolator = Extrapolator(
            PchipInterpolator(np.array([3, 4, 5]), np.array([1, 2, 3]))
        )
        np.testing.assert_almost_equal(
            extrapolator((0.1, 0.2, 8.0, 9.0)), (-1.9, -1.8, 6.0, 7.0)
        )
        self.assertEqual(extrapolator(9), 7.0)

    @ignore_numpy_errors
    def test_nan__call__(self):
        """
        Test :method:`colour.algebra.extrapolation.Extrapolator.__call__`
        method nan support.
        """

        cases = [-1.0, 0.0, 1.0, -np.inf, np.inf, np.nan]
        cases = set(permutations(cases * 3, r=3))
        for case in cases:
            extrapolator = Extrapolator(
                LinearInterpolator(np.array(case), np.array(case))
            )
            extrapolator(case[0])


if __name__ == "__main__":
    unittest.main()



"""
Colour Matching Functions Transformations
=========================================

Defines various educational objects for colour matching functions
transformations:

-   :func:`colour.colorimetry.RGB_2_degree_cmfs_to_XYZ_2_degree_cmfs`
-   :func:`colour.colorimetry.RGB_10_degree_cmfs_to_XYZ_10_degree_cmfs`
-   :func:`colour.colorimetry.RGB_10_degree_cmfs_to_LMS_10_degree_cmfs`
-   :func:`colour.colorimetry.LMS_2_degree_cmfs_to_XYZ_2_degree_cmfs`
-   :func:`colour.colorimetry.LMS_10_degree_cmfs_to_XYZ_10_degree_cmfs`

References
----------
-   :cite:`CIETC1-362006a` : CIE TC 1-36. (2006). CIE 170-1:2006 Fundamental
    Chromaticity Diagram with Physiological Axes - Part 1. Commission
    Internationale de l'Eclairage. ISBN:978-3-901906-46-6
-   :cite:`CVRLp` : CVRL. (n.d.). CIE (2012) 10-deg XYZ
    "physiologically-relevant" colour matching functions. Retrieved June 25,
    2014, from http://www.cvrl.org/database/text/cienewxyz/cie2012xyz10.htm
-   :cite:`CVRLv` : CVRL. (n.d.). CIE (2012) 2-deg XYZ
    "physiologically-relevant" colour matching functions. Retrieved June 25,
    2014, from http://www.cvrl.org/database/text/cienewxyz/cie2012xyz2.htm
-   :cite:`Wyszecki2000be` : Wyszecki, Günther, & Stiles, W. S. (2000). The
    CIE 1964 Standard Observer. In Color Science: Concepts and Methods,
    Quantitative Data and Formulae (p. 141). Wiley. ISBN:978-0-471-39918-6
-   :cite:`Wyszecki2000bg` : Wyszecki, Günther, & Stiles, W. S. (2000). Table
    1(3.3.3). In Color Science: Concepts and Methods, Quantitative Data and
    Formulae (pp. 138-139). Wiley. ISBN:978-0-471-39918-6
"""

from __future__ import annotations

import numpy as np

from colour.algebra import vector_dot
from colour.colorimetry import (
    MSDS_CMFS_LMS,
    MSDS_CMFS_RGB,
    SDS_LEFS_PHOTOPIC,
    reshape_sd,
)
from colour.hints import FloatingOrArrayLike, NDArray
from colour.utilities import tstack

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "RGB_2_degree_cmfs_to_XYZ_2_degree_cmfs",
    "RGB_10_degree_cmfs_to_XYZ_10_degree_cmfs",
    "RGB_10_degree_cmfs_to_LMS_10_degree_cmfs",
    "LMS_2_degree_cmfs_to_XYZ_2_degree_cmfs",
    "LMS_10_degree_cmfs_to_XYZ_10_degree_cmfs",
]


def RGB_2_degree_cmfs_to_XYZ_2_degree_cmfs(
    wavelength: FloatingOrArrayLike,
) -> NDArray:
    """
    Convert *Wright & Guild 1931 2 Degree RGB CMFs* colour matching functions
    into the *CIE 1931 2 Degree Standard Observer* colour matching functions.

    Parameters
    ----------
    wavelength
        Wavelength :math:`\\lambda` in nm.

    Returns
    -------
    :class:`numpy.ndarray`
        *CIE 1931 2 Degree Standard Observer* spectral tristimulus values.

    Notes
    -----
    -   Data for the *CIE 1931 2 Degree Standard Observer* already exists,
        this definition is intended for educational purpose.

    References
    ----------
    :cite:`Wyszecki2000bg`

    Examples
    --------
    >>> from colour.utilities import numpy_print_options
    >>> with numpy_print_options(suppress=True):
    ...     RGB_2_degree_cmfs_to_XYZ_2_degree_cmfs(700)  # doctest: +ELLIPSIS
    array([ 0.0113577...,  0.004102  ,  0.        ])
    """

    cmfs = MSDS_CMFS_RGB["Wright & Guild 1931 2 Degree RGB CMFs"]

    rgb_bar = cmfs[wavelength]

    rgb = rgb_bar / np.sum(rgb_bar)

    M1 = np.array(
        [
            [0.49000, 0.31000, 0.20000],
            [0.17697, 0.81240, 0.01063],
            [0.00000, 0.01000, 0.99000],
        ]
    )

    M2 = np.array(
        [
            [0.66697, 1.13240, 1.20063],
            [0.66697, 1.13240, 1.20063],
            [0.66697, 1.13240, 1.20063],
        ]
    )

    xyz = vector_dot(M1, rgb)
    xyz /= vector_dot(M2, rgb)

    x, y, z = xyz[..., 0], xyz[..., 1], xyz[..., 2]

    V = reshape_sd(
        SDS_LEFS_PHOTOPIC["CIE 1924 Photopic Standard Observer"], cmfs.shape
    )
    L = V[wavelength]

    x_bar = x / y * L
    y_bar = L
    z_bar = z / y * L

    xyz_bar = tstack([x_bar, y_bar, z_bar])

    return xyz_bar


def RGB_10_degree_cmfs_to_XYZ_10_degree_cmfs(
    wavelength: FloatingOrArrayLike,
) -> NDArray:
    """
    Convert *Stiles & Burch 1959 10 Degree RGB CMFs* colour matching
    functions into the *CIE 1964 10 Degree Standard Observer* colour matching
    functions.

    Parameters
    ----------
    wavelength
        Wavelength :math:`\\lambda` in nm.

    Returns
    -------
    :class:`numpy.ndarray`
        *CIE 1964 10 Degree Standard Observer* spectral tristimulus values.

    Notes
    -----
    -   Data for the *CIE 1964 10 Degree Standard Observer* already exists,
        this definition is intended for educational purpose.

    References
    ----------
    :cite:`Wyszecki2000be`

    Examples
    --------
    >>> from colour.utilities import numpy_print_options
    >>> with numpy_print_options(suppress=True):
    ...     RGB_10_degree_cmfs_to_XYZ_10_degree_cmfs(700)  # doctest: +ELLIPSIS
    array([ 0.0096432...,  0.0037526..., -0.0000041...])
    """

    cmfs = MSDS_CMFS_RGB["Stiles & Burch 1959 10 Degree RGB CMFs"]

    rgb_bar = cmfs[wavelength]

    M = np.array(
        [
            [0.341080, 0.189145, 0.387529],
            [0.139058, 0.837460, 0.073316],
            [0.000000, 0.039553, 2.026200],
        ]
    )

    xyz_bar = vector_dot(M, rgb_bar)

    return xyz_bar


def RGB_10_degree_cmfs_to_LMS_10_degree_cmfs(
    wavelength: FloatingOrArrayLike,
) -> NDArray:
    """
    Convert *Stiles & Burch 1959 10 Degree RGB CMFs* colour matching
    functions into the *Stockman & Sharpe 10 Degree Cone Fundamentals*
    spectral sensitivity functions.

    Parameters
    ----------
    wavelength
        Wavelength :math:`\\lambda` in nm.

    Returns
    -------
    :class:`numpy.ndarray`
        *Stockman & Sharpe 10 Degree Cone Fundamentals* spectral tristimulus
        values.

    Notes
    -----
    -   Data for the *Stockman & Sharpe 10 Degree Cone Fundamentals* already
        exists, this definition is intended for educational purpose.

    References
    ----------
    :cite:`CIETC1-362006a`

    Examples
    --------
    >>> from colour.utilities import numpy_print_options
    >>> with numpy_print_options(suppress=True):
    ...     RGB_10_degree_cmfs_to_LMS_10_degree_cmfs(700)  # doctest: +ELLIPSIS
    array([ 0.0052860...,  0.0003252...,  0.        ])
    """

    cmfs = MSDS_CMFS_RGB["Stiles & Burch 1959 10 Degree RGB CMFs"]

    rgb_bar = cmfs[wavelength]

    M = np.array(
        [
            [0.1923252690, 0.749548882, 0.0675726702],
            [0.0192290085, 0.940908496, 0.113830196],
            [0.0000000000, 0.0105107859, 0.991427669],
        ]
    )

    lms_bar = vector_dot(M, rgb_bar)
    lms_bar[..., -1][np.asarray(np.asarray(wavelength) > 505)] = 0

    return lms_bar


def LMS_2_degree_cmfs_to_XYZ_2_degree_cmfs(
    wavelength: FloatingOrArrayLike,
) -> NDArray:
    """
    Convert *Stockman & Sharpe 2 Degree Cone Fundamentals* colour matching
    functions into the *CIE 2012 2 Degree Standard Observer* colour matching
    functions.

    Parameters
    ----------
    wavelength
        Wavelength :math:`\\lambda` in nm.

    Returns
    -------
    :class:`numpy.ndarray`
        *CIE 2012 2 Degree Standard Observer* spectral tristimulus values.

    Notes
    -----
    -   Data for the *CIE 2012 2 Degree Standard Observer* already exists,
        this definition is intended for educational purpose.

    References
    ----------
    :cite:`CVRLv`

    Examples
    --------
    >>> from colour.utilities import numpy_print_options
    >>> with numpy_print_options(suppress=True):
    ...     LMS_2_degree_cmfs_to_XYZ_2_degree_cmfs(700)  # doctest: +ELLIPSIS
    array([ 0.0109677...,  0.0041959...,  0.        ])
    """

    cmfs = MSDS_CMFS_LMS["Stockman & Sharpe 2 Degree Cone Fundamentals"]

    lms_bar = cmfs[wavelength]

    M = np.array(
        [
            [1.94735469, -1.41445123, 0.36476327],
            [0.68990272, 0.34832189, 0.00000000],
            [0.00000000, 0.00000000, 1.93485343],
        ]
    )

    xyz_bar = vector_dot(M, lms_bar)

    return xyz_bar


def LMS_10_degree_cmfs_to_XYZ_10_degree_cmfs(
    wavelength: FloatingOrArrayLike,
) -> NDArray:
    """
    Convert *Stockman & Sharpe 10 Degree Cone Fundamentals* colour matching
    functions into the *CIE 2012 10 Degree Standard Observer* colour matching
    functions.

    Parameters
    ----------
    wavelength
        Wavelength :math:`\\lambda` in nm.

    Returns
    -------
    :class:`numpy.ndarray`
        *CIE 2012 10 Degree Standard Observer* spectral tristimulus values.

    Notes
    -----
    -   Data for the *CIE 2012 10 Degree Standard Observer* already exists,
        this definition is intended for educational purpose.

    References
    ----------
    :cite:`CVRLp`

    Examples
    --------
    >>> from colour.utilities import numpy_print_options
    >>> with numpy_print_options(suppress=True):
    ...     LMS_10_degree_cmfs_to_XYZ_10_degree_cmfs(700)  # doctest: +ELLIPSIS
    array([ 0.0098162...,  0.0037761...,  0.        ])
    """

    cmfs = MSDS_CMFS_LMS["Stockman & Sharpe 10 Degree Cone Fundamentals"]

    lms_bar = cmfs[wavelength]

    M = np.array(
        [
            [1.93986443, -1.34664359, 0.43044935],
            [0.69283932, 0.34967567, 0.00000000],
            [0.00000000, 0.00000000, 2.14687945],
        ]
    )

    xyz_bar = vector_dot(M, lms_bar)

    return xyz_bar



"""Showcases reflectance recovery computations using *Meng et al. (2015)* method."""

import numpy as np

import colour
from colour.utilities import message_box

message_box('"Meng et al. (2015)" - Reflectance Recovery Computations')

illuminant = colour.SDS_ILLUMINANTS["D65"]

XYZ = np.array([0.20654008, 0.12197225, 0.05136952])
message_box(
    f'Recovering reflectance using "Meng et al. (2015)" method from given '
    f'"XYZ" tristimulus values:\n\n\tXYZ: {XYZ}'
)
sd = colour.XYZ_to_sd(XYZ, method="Meng 2015")
print(sd)
print(colour.recovery.XYZ_to_sd_Meng2015(XYZ))
print(colour.sd_to_XYZ(sd, illuminant=illuminant) / 100)



"""
SRGB
====

Defines the *sRGB* electro-optical transfer function (EOTF) and its
inverse:

-   :func:`colour.models.eotf_inverse_sRGB`
-   :func:`colour.models.eotf_sRGB`

References
----------
-   :cite:`InternationalElectrotechnicalCommission1999a` : International
    Electrotechnical Commission. (1999). IEC 61966-2-1:1999 - Multimedia
    systems and equipment - Colour measurement and management - Part 2-1:
    Colour management - Default RGB colour space - sRGB (p. 51).
    https://webstore.iec.ch/publication/6169
-   :cite:`InternationalTelecommunicationUnion2015i` : International
    Telecommunication Union. (2015). Recommendation ITU-R BT.709-6 - Parameter
    values for the HDTV standards for production and international programme
    exchange BT Series Broadcasting service (pp. 1-32).
    https://www.itu.int/dms_pubrec/itu-r/rec/bt/\
R-REC-BT.709-6-201506-I!!PDF-E.pdf
"""

from __future__ import annotations

import numpy as np

from colour.algebra import spow
from colour.hints import FloatingOrArrayLike, FloatingOrNDArray
from colour.utilities import (
    as_float,
    domain_range_scale,
    from_range_1,
    to_domain_1,
)

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "eotf_inverse_sRGB",
    "eotf_sRGB",
]


def eotf_inverse_sRGB(L: FloatingOrArrayLike) -> FloatingOrNDArray:
    """
    Define the *IEC 61966-2-1:1999* *sRGB* inverse electro-optical transfer
    function (EOTF).

    Parameters
    ----------
    L
        *Luminance* :math:`L` of the image.

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        Corresponding electrical signal :math:`V`.

    Notes
    -----
    +------------+-----------------------+---------------+
    | **Domain** | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``L``      | [0, 1]                | [0, 1]        |
    +------------+-----------------------+---------------+

    +------------+-----------------------+---------------+
    | **Range**  | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``V``      | [0, 1]                | [0, 1]        |
    +------------+-----------------------+---------------+

    References
    ----------
    :cite:`InternationalElectrotechnicalCommission1999a`,
    :cite:`InternationalTelecommunicationUnion2015i`

    Examples
    --------
    >>> eotf_inverse_sRGB(0.18)  # doctest: +ELLIPSIS
    0.4613561...
    """

    L = to_domain_1(L)

    V = np.where(L <= 0.0031308, L * 12.92, 1.055 * spow(L, 1 / 2.4) - 0.055)

    return as_float(from_range_1(V))


def eotf_sRGB(V: FloatingOrArrayLike) -> FloatingOrNDArray:
    """
    Define the *IEC 61966-2-1:1999* *sRGB* electro-optical transfer function
    (EOTF).

    Parameters
    ----------
    V
        Electrical signal :math:`V`.

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        Corresponding *luminance* :math:`L` of the image.

    Notes
    -----
    +------------+-----------------------+---------------+
    | **Domain** | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``V``      | [0, 1]                | [0, 1]        |
    +------------+-----------------------+---------------+

    +------------+-----------------------+---------------+
    | **Range**  | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``L``      | [0, 1]                | [0, 1]        |
    +------------+-----------------------+---------------+

    References
    ----------
    :cite:`InternationalElectrotechnicalCommission1999a`,
    :cite:`InternationalTelecommunicationUnion2015i`

    Examples
    --------
    >>> eotf_sRGB(0.461356129500442)  # doctest: +ELLIPSIS
    0.1...
    """

    V = to_domain_1(V)

    with domain_range_scale("ignore"):
        L = np.where(
            V <= eotf_inverse_sRGB(0.0031308),
            V / 12.92,
            spow((V + 0.055) / 1.055, 2.4),
        )

    return as_float(from_range_1(L))



"""
Colour Notation Systems Plotting
================================

Defines the colour notation systems plotting objects:

-   :func:`colour.plotting.plot_single_munsell_value_function`
-   :func:`colour.plotting.plot_multi_munsell_value_functions`
"""

from __future__ import annotations

import matplotlib.pyplot as plt
import numpy as np

from colour.hints import Any, Callable, Dict, Sequence, Tuple, Union
from colour.notation import MUNSELL_VALUE_METHODS
from colour.plotting import (
    filter_passthrough,
    plot_multi_functions,
    override_style,
)

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "plot_single_munsell_value_function",
    "plot_multi_munsell_value_functions",
]


@override_style()
def plot_single_munsell_value_function(
    function: Union[Callable, str], **kwargs: Any
) -> Tuple[plt.Figure, plt.Axes]:
    """
    Plot given *Lightness* function.

    Parameters
    ----------
    function
        *Munsell* value function to plot. ``function`` can be of any type or
        form supported by the :func:`colour.plotting.filter_passthrough`
        definition.

    Other Parameters
    ----------------
    kwargs
        {:func:`colour.plotting.artist`,
        :func:`colour.plotting.plot_multi_functions`,
        :func:`colour.plotting.render`},
        See the documentation of the previously listed definitions.

    Returns
    -------
    :class:`tuple`
        Current figure and axes.

    Examples
    --------
    >>> plot_single_munsell_value_function('ASTM D1535')  # doctest: +ELLIPSIS
    (<Figure size ... with 1 Axes>, <...AxesSubplot...>)

    .. image:: ../_static/Plotting_Plot_Single_Munsell_Value_Function.png
        :align: center
        :alt: plot_single_munsell_value_function
    """

    settings: Dict[str, Any] = {
        "title": f"{function} - Munsell Value Function"
    }
    settings.update(kwargs)

    return plot_multi_munsell_value_functions((function,), **settings)


@override_style()
def plot_multi_munsell_value_functions(
    functions: Union[Callable, str, Sequence[Union[Callable, str]]],
    **kwargs: Any,
) -> Tuple[plt.Figure, plt.Axes]:
    """
    Plot given *Munsell* value functions.

    Parameters
    ----------
    functions
        *Munsell* value functions to plot. ``functions`` elements can be of any
        type or form supported by the
        :func:`colour.plotting.filter_passthrough` definition.

    Other Parameters
    ----------------
    kwargs
        {:func:`colour.plotting.artist`,
        :func:`colour.plotting.plot_multi_functions`,
        :func:`colour.plotting.render`},
        See the documentation of the previously listed definitions.

    Returns
    -------
    :class:`tuple`
        Current figure and axes.

    Examples
    --------
    >>> plot_multi_munsell_value_functions(['ASTM D1535', 'McCamy 1987'])
    ... # doctest: +ELLIPSIS
    (<Figure size ... with 1 Axes>, <...AxesSubplot...>)

    .. image:: ../_static/Plotting_Plot_Multi_Munsell_Value_Functions.png
        :align: center
        :alt: plot_multi_munsell_value_functions
    """

    functions_filtered = filter_passthrough(MUNSELL_VALUE_METHODS, functions)

    settings: Dict[str, Any] = {
        "bounding_box": (0, 100, 0, 10),
        "legend": True,
        "title": f"{', '.join(functions_filtered)} - Munsell Functions",
        "x_label": "Luminance Y",
        "y_label": "Munsell Value V",
    }
    settings.update(kwargs)

    return plot_multi_functions(
        functions_filtered, samples=np.linspace(0, 100, 1000), **settings
    )



"""
Defines the unit tests for the :mod:`colour.models.rgb.transfer_functions.\
blackmagic_design` module.
"""

import numpy as np
import unittest

from colour.models.rgb.transfer_functions import (
    oetf_BlackmagicFilmGeneration5,
    oetf_inverse_BlackmagicFilmGeneration5,
)
from colour.utilities import domain_range_scale, ignore_numpy_errors

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "TestOetf_BlackmagicFilmGeneration5",
    "TestOetf_inverse_BlackmagicFilmGeneration5",
]


class TestOetf_BlackmagicFilmGeneration5(unittest.TestCase):
    """
    Define :func:`colour.models.rgb.transfer_functions.blackmagic_design.\
oetf_BlackmagicFilmGeneration5` definition unit tests methods.
    """

    def test_oetf_BlackmagicFilmGeneration5(self):
        """
        Test :func:`colour.models.rgb.transfer_functions.\
blackmagic_design.oetf_BlackmagicFilmGeneration5` definition.
        """

        self.assertAlmostEqual(
            oetf_BlackmagicFilmGeneration5(0.0), 0.092465753424658, places=7
        )

        self.assertAlmostEqual(
            oetf_BlackmagicFilmGeneration5(0.18), 0.383561643835617, places=7
        )

        self.assertAlmostEqual(
            oetf_BlackmagicFilmGeneration5(1.0), 0.530489624957305, places=7
        )

        self.assertAlmostEqual(
            oetf_BlackmagicFilmGeneration5(100.0), 0.930339851899973, places=7
        )

        self.assertAlmostEqual(
            oetf_BlackmagicFilmGeneration5(222.86), 0.999999631713769, places=7
        )

    def test_n_dimensional_oetf_BlackmagicFilmGeneration5(self):
        """
        Test :func:`colour.models.rgb.transfer_functions.\
blackmagic_design.oetf_BlackmagicFilmGeneration5` definition n-dimensional
        arrays support.
        """

        L = 0.18
        V = oetf_BlackmagicFilmGeneration5(L)

        L = np.tile(L, 6)
        V = np.tile(V, 6)
        np.testing.assert_almost_equal(
            oetf_BlackmagicFilmGeneration5(L), V, decimal=7
        )

        L = np.reshape(L, (2, 3))
        V = np.reshape(V, (2, 3))
        np.testing.assert_almost_equal(
            oetf_BlackmagicFilmGeneration5(L), V, decimal=7
        )

        L = np.reshape(L, (2, 3, 1))
        V = np.reshape(V, (2, 3, 1))
        np.testing.assert_almost_equal(
            oetf_BlackmagicFilmGeneration5(L), V, decimal=7
        )

    def test_domain_range_scale_oetf_BlackmagicFilmGeneration5(self):
        """
        Test :func:`colour.models.rgb.transfer_functions.\
blackmagic_design.oetf_BlackmagicFilmGeneration5` definition domain and range
        scale support.
        """

        L = 0.18
        V = oetf_BlackmagicFilmGeneration5(L)

        d_r = (("reference", 1), ("1", 1), ("100", 100))
        for scale, factor in d_r:
            with domain_range_scale(scale):
                np.testing.assert_almost_equal(
                    oetf_BlackmagicFilmGeneration5(L * factor),
                    V * factor,
                    decimal=7,
                )

    @ignore_numpy_errors
    def test_nan_oetf_BlackmagicFilmGeneration5(self):
        """
        Test :func:`colour.models.rgb.transfer_functions.\
blackmagic_design.oetf_BlackmagicFilmGeneration5` definition nan support.
        """

        oetf_BlackmagicFilmGeneration5(
            np.array([-1.0, 0.0, 1.0, -np.inf, np.inf, np.nan])
        )


class TestOetf_inverse_BlackmagicFilmGeneration5(unittest.TestCase):
    """
    Define :func:`colour.models.rgb.transfer_functions.\
blackmagic_design.oetf_inverse_BlackmagicFilmGeneration5` definition unit tests
    methods.
    """

    def test_oetf_inverse_BlackmagicFilmGeneration5(self):
        """
        Test :func:`colour.models.rgb.transfer_functions.\
blackmagic_design.oetf_inverse_BlackmagicFilmGeneration5` definition.
        """

        self.assertAlmostEqual(
            oetf_inverse_BlackmagicFilmGeneration5(0.092465753424658),
            0.0,
            places=7,
        )

        self.assertAlmostEqual(
            oetf_inverse_BlackmagicFilmGeneration5(0.383561643835617),
            0.18,
            places=7,
        )

        self.assertAlmostEqual(
            oetf_inverse_BlackmagicFilmGeneration5(0.530489624957305),
            1.0,
            places=7,
        )

        self.assertAlmostEqual(
            oetf_inverse_BlackmagicFilmGeneration5(0.930339851899973),
            100.0,
            places=7,
        )

        self.assertAlmostEqual(
            oetf_inverse_BlackmagicFilmGeneration5(0.999999631713769),
            222.86,
            places=7,
        )

    def test_n_dimensional_oetf_inverse_BlackmagicFilmGeneration5(self):
        """
        Test :func:`colour.models.rgb.transfer_functions.\
blackmagic_design.oetf_inverse_BlackmagicFilmGeneration5` definition
        n-dimensional arrays support.
        """

        V = 0.383561643835617
        L = oetf_inverse_BlackmagicFilmGeneration5(V)

        V = np.tile(V, 6)
        L = np.tile(L, 6)
        np.testing.assert_almost_equal(
            oetf_inverse_BlackmagicFilmGeneration5(V), L, decimal=7
        )

        V = np.reshape(V, (2, 3))
        L = np.reshape(L, (2, 3))
        np.testing.assert_almost_equal(
            oetf_inverse_BlackmagicFilmGeneration5(V), L, decimal=7
        )

        V = np.reshape(V, (2, 3, 1))
        L = np.reshape(L, (2, 3, 1))
        np.testing.assert_almost_equal(
            oetf_inverse_BlackmagicFilmGeneration5(V), L, decimal=7
        )

    def test_domain_range_scale_oetf_inverse_BlackmagicFilmGeneration5(self):
        """
        Test :func:`colour.models.rgb.transfer_functions.\
blackmagic_design.oetf_inverse_BlackmagicFilmGeneration5` definition domain and
        range scale support.
        """

        V = 0.383561643835617
        L = oetf_inverse_BlackmagicFilmGeneration5(V)

        d_r = (("reference", 1), ("1", 1), ("100", 100))
        for scale, factor in d_r:
            with domain_range_scale(scale):
                np.testing.assert_almost_equal(
                    oetf_inverse_BlackmagicFilmGeneration5(V * factor),
                    L * factor,
                    decimal=7,
                )

    @ignore_numpy_errors
    def test_nan_oetf_inverse_BlackmagicFilmGeneration5(self):
        """
        Test :func:`colour.models.rgb.transfer_functions.\
blackmagic_design.oetf_inverse_BlackmagicFilmGeneration5` definition nan
        support.
        """

        oetf_inverse_BlackmagicFilmGeneration5(
            np.array([-1.0, 0.0, 1.0, -np.inf, np.inf, np.nan])
        )


if __name__ == "__main__":
    unittest.main()



"""
ITU-R BT.709-6
==============

Defines the *ITU-R BT.709-6* opto-electrical transfer function (OETF)
and its inverse:

-   :func:`colour.models.oetf_BT709`
-   :func:`colour.models.oetf_inverse_BT709`

References
----------
-   :cite:`InternationalTelecommunicationUnion2015i` : International
    Telecommunication Union. (2015). Recommendation ITU-R BT.709-6 - Parameter
    values for the HDTV standards for production and international programme
    exchange BT Series Broadcasting service (pp. 1-32).
    https://www.itu.int/dms_pubrec/itu-r/rec/bt/\
R-REC-BT.709-6-201506-I!!PDF-E.pdf
"""

from __future__ import annotations

from colour.hints import FloatingOrArrayLike, FloatingOrNDArray
from colour.models.rgb.transfer_functions import oetf_BT601, oetf_inverse_BT601

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "oetf_BT709",
    "oetf_inverse_BT709",
]


def oetf_BT709(L: FloatingOrArrayLike) -> FloatingOrNDArray:
    """
    Define *Recommendation ITU-R BT.709-6* opto-electronic transfer function
    (OETF).

    Parameters
    ----------
    L
        *Luminance* :math:`L` of the image.

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        Corresponding electrical signal :math:`V`.

    Notes
    -----
    +------------+-----------------------+---------------+
    | **Domain** | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``L``      | [0, 1]                | [0, 1]        |
    +------------+-----------------------+---------------+

    +------------+-----------------------+---------------+
    | **Range**  | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``V``      | [0, 1]                | [0, 1]        |
    +------------+-----------------------+---------------+

    References
    ----------
    :cite:`InternationalTelecommunicationUnion2015i`

    Examples
    --------
    >>> oetf_BT709(0.18)  # doctest: +ELLIPSIS
    0.4090077...
    """

    return oetf_BT601(L)


def oetf_inverse_BT709(V: FloatingOrArrayLike) -> FloatingOrNDArray:
    """
    Define *Recommendation ITU-R BT.709-6* inverse opto-electronic transfer
    function (OETF).

    Parameters
    ----------
    V
        Electrical signal :math:`V`.

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        Corresponding *luminance* :math:`L` of the image.

    Notes
    -----
    +------------+-----------------------+---------------+
    | **Domain** | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``V``      | [0, 1]                | [0, 1]        |
    +------------+-----------------------+---------------+

    +------------+-----------------------+---------------+
    | **Range**  | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``L``      | [0, 1]                | [0, 1]        |
    +------------+-----------------------+---------------+

    References
    ----------
    :cite:`InternationalTelecommunicationUnion2015i`

    Examples
    --------
    >>> oetf_inverse_BT709(0.409007728864150)  # doctest: +ELLIPSIS
    0.1...
    """

    return oetf_inverse_BT601(V)



"""
Huang et al. (2015) Power-Functions
===================================

Defines the *Huang, Cui, Melgosa, Sanchez-Maranon, Li, Luo and Liu (2015)*
power-functions improving the performance of colour-difference formulas:

-   :func:`colour.difference.power_function_Huang2015`

References
----------
-   :cite:`Huang2015` : Huang, M., Cui, G., Melgosa, M., Sanchez-Maranon, M.,
    Li, C., Luo, M. R., & Liu, H. (2015). Power functions improving the
    performance of color-difference formulas. Optical Society of America,
    23(1), 597-610. doi:10.1364/OE.23.000597
-   :cite:`Li2017` : Li, C., Li, Z., Wang, Z., Xu, Y., Luo, M. R., Cui, G.,
    Melgosa, M., Brill, M. H., & Pointer, M. (2017). Comprehensive color
    solutions: CAM16, CAT16, and CAM16-UCS. Color Research & Application,
    42(6), 703-718. doi:10.1002/col.22131
"""

from __future__ import annotations

import numpy as np

from colour.hints import FloatingOrArrayLike, FloatingOrNDArray, Literal, Union
from colour.utilities import CaseInsensitiveMapping, tsplit, validate_method

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "power_function_Huang2015",
]

COEFFICIENTS_HUANG2015: CaseInsensitiveMapping = CaseInsensitiveMapping(
    {
        "CIE 1976": np.array([1.26, 0.55]),
        "CIE 1994": np.array([1.41, 0.70]),
        "CIE 2000": np.array([1.43, 0.70]),
        "CMC": np.array([1.34, 0.66]),
        "CAM02-LCD": np.array([1.00, 0.85]),
        "CAM02-SCD": np.array([1.45, 0.75]),
        "CAM02-UCS": np.array([1.30, 0.75]),
        "CAM16-UCS": np.array([1.41, 0.63]),
        "DIN99d": np.array([1.28, 0.74]),
        "OSA": np.array([3.32, 0.62]),
        "OSA-GP-Euclidean": np.array([1.52, 0.76]),
        "ULAB": np.array([1.17, 0.69]),
    }
)
COEFFICIENTS_HUANG2015.__doc__ = """
*Huang et al. (2015)* power-functions coefficients.

References
----------
:cite:`Huang2015`, :cite:`Li2017`

Notes
-----
-   :cite:`Li2017` does not give the coefficients for the *CAM16-LCD* and
    *CAM16-SCD* colourspaces. *Ronnie Luo* has been contacted to know if they
    have been computed.

Aliases:

-   'cie1976': 'CIE 1976'
-   'cie1994': 'CIE 1994'
-   'cie2000': 'CIE 2000'
"""
COEFFICIENTS_HUANG2015["cie1976"] = COEFFICIENTS_HUANG2015["CIE 1976"]
COEFFICIENTS_HUANG2015["cie1994"] = COEFFICIENTS_HUANG2015["CIE 1994"]
COEFFICIENTS_HUANG2015["cie2000"] = COEFFICIENTS_HUANG2015["CIE 2000"]


def power_function_Huang2015(
    d_E: FloatingOrArrayLike,
    coefficients: Union[
        Literal[
            "CIE 1976",
            "CIE 1994",
            "CIE 2000",
            "CMC",
            "CAM02-LCD",
            "CAM02-SCD",
            "CAM16-UCS",
            "DIN99d",
            "OSA",
            "OSA-GP-Euclidean",
            "ULAB",
        ],
        str,
    ] = "CIE 2000",
) -> FloatingOrNDArray:
    """
    Improve the performance of the :math:`\\Delta E` value for given
    coefficients using
    *Huang, Cui, Melgosa, Sanchez-Maranon, Li, Luo and Liu (2015)*
    power-function: :math:`d_E^{\\prime}=a*d_{E^b}`.

    Parameters
    ----------
    d_E
        Computed colour difference array :math:`\\Delta E`.
    coefficients
        Coefficients for the power-function.

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        Improved math:`\\Delta E` value.

    References
    ----------
    :cite:`Huang2015`, :cite:`Li2017`

    Examples
    --------
    >>> d_E = np.array([2.0425, 2.8615, 3.4412])
    >>> power_function_Huang2015(d_E)  # doctest: +ELLIPSIS
    array([ 2.3574879...,  2.9850503...,  3.3965106...])
    """

    coefficients = validate_method(
        coefficients,
        COEFFICIENTS_HUANG2015,
        '"{0}" coefficients are invalid, ' "they must be one of {1}!",
    )

    a, b = tsplit(COEFFICIENTS_HUANG2015[coefficients])

    d_E_p = a * d_E**b

    return d_E_p



"""
Cameras Sensitivities
=====================

Defines the spectral distributions classes for the datasets from
the :mod:`colour.characterisation.datasets.cameras` module:

-   :class:`colour.characterisation.RGB_CameraSensitivities`: Implements
    support for a camera *RGB* sensitivities.
"""

from __future__ import annotations

from colour.colorimetry import (
    MultiSpectralDistributions,
    SpectralDistribution,
    SpectralShape,
)
from colour.continuous import MultiSignals, Signal
from colour.hints import ArrayLike, Any, Optional, Sequence, Union
from colour.utilities import is_pandas_installed

if is_pandas_installed():
    from pandas import DataFrame, Series
else:  # pragma: no cover
    from unittest import mock

    DataFrame = mock.MagicMock()
    Series = mock.MagicMock()

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "RGB_CameraSensitivities",
]


class RGB_CameraSensitivities(MultiSpectralDistributions):
    """
    Implement support for a camera *RGB* sensitivities.

    Parameters
    ----------
    data
        Data to be stored in the multi-spectral distributions.
    domain
        Values to initialise the multiple :class:`colour.SpectralDistribution`
        class instances :attr:`colour.continuous.Signal.wavelengths` attribute
        with. If both ``data`` and ``domain`` arguments are defined, the latter
        will be used to initialise the
        :attr:`colour.continuous.Signal.wavelengths` property.
    labels
        Names to use for the :class:`colour.SpectralDistribution` class
        instances.

    Other Parameters
    ----------------
    extrapolator
        Extrapolator class type to use as extrapolating function for the
        :class:`colour.SpectralDistribution` class instances.
    extrapolator_kwargs
        Arguments to use when instantiating the extrapolating function
        of the :class:`colour.SpectralDistribution` class instances.
    interpolator
        Interpolator class type to use as interpolating function for the
        :class:`colour.SpectralDistribution` class instances.
    interpolator_kwargs
        Arguments to use when instantiating the interpolating function
        of the :class:`colour.SpectralDistribution` class instances.
    name
       Multi-spectral distributions name.
    strict_labels
        Multi-spectral distributions labels for figures, default to
        :attr:`colour.colorimetry.RGB_CameraSensitivities.labels` property
        value.
    """

    def __init__(
        self,
        data: Optional[
            Union[
                ArrayLike,
                DataFrame,
                dict,
                MultiSignals,
                MultiSpectralDistributions,
                Sequence,
                Series,
                Signal,
                SpectralDistribution,
            ]
        ] = None,
        domain: Optional[Union[ArrayLike, SpectralShape]] = None,
        labels: Optional[Sequence] = None,
        **kwargs: Any,
    ):
        super().__init__(
            data, domain, labels=("red", "green", "blue"), **kwargs
        )



from .munsell import (
    MUNSELL_COLOURS_ALL,
    MUNSELL_COLOURS_1929,
    MUNSELL_COLOURS_REAL,
    MUNSELL_COLOURS,
)

__all__ = [
    "MUNSELL_COLOURS_ALL",
    "MUNSELL_COLOURS_1929",
    "MUNSELL_COLOURS_REAL",
    "MUNSELL_COLOURS",
]



"""Defines the unit tests for the :mod:`colour.plotting.quality` module."""

import unittest
from matplotlib.pyplot import Axes, Figure

from colour.colorimetry import (
    SDS_ILLUMINANTS,
    SDS_LIGHT_SOURCES,
    SpectralShape,
    reshape_sd,
)
from colour.plotting import (
    plot_single_sd_colour_rendering_index_bars,
    plot_multi_sds_colour_rendering_indexes_bars,
    plot_single_sd_colour_quality_scale_bars,
    plot_multi_sds_colour_quality_scales_bars,
)
from colour.plotting.quality import plot_colour_quality_bars
from colour.quality import colour_quality_scale

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "TestPlotColourQualityBars",
    "TestPlotSingleSdColourRenderingIndexBars",
    "TestPlotMultiSdsColourRenderingIndexesBars",
    "TestPlotSingleSdColourQualityScaleBars",
    "TestPlotMultiSdsColourQualityScalesBars",
]


class TestPlotColourQualityBars(unittest.TestCase):
    """
    Define :func:`colour.plotting.quality.plot_colour_quality_bars` definition
    unit tests methods.
    """

    def test_plot_colour_quality_bars(self):
        """
        Test :func:`colour.plotting.quality.plot_colour_quality_bars`
        definition.
        """

        illuminant = SDS_ILLUMINANTS["FL2"]
        light_source = SDS_LIGHT_SOURCES["Kinoton 75P"]
        light_source = reshape_sd(light_source, SpectralShape(360, 830, 1))
        cqs_i = colour_quality_scale(illuminant, additional_data=True)
        cqs_l = colour_quality_scale(light_source, additional_data=True)

        figure, axes = plot_colour_quality_bars([cqs_i, cqs_l])

        self.assertIsInstance(figure, Figure)
        self.assertIsInstance(axes, Axes)


class TestPlotSingleSdColourRenderingIndexBars(unittest.TestCase):
    """
    Define :func:`colour.plotting.quality.\
plot_single_sd_colour_rendering_index_bars` definition unit tests methods.
    """

    def test_plot_single_sd_colour_rendering_index_bars(self):
        """
        Test :func:`colour.plotting.quality.\
plot_single_sd_colour_rendering_index_bars` definition.
        """

        figure, axes = plot_single_sd_colour_rendering_index_bars(
            SDS_ILLUMINANTS["FL2"]
        )

        self.assertIsInstance(figure, Figure)
        self.assertIsInstance(axes, Axes)


class TestPlotMultiSdsColourRenderingIndexesBars(unittest.TestCase):
    """
    Define :func:`colour.plotting.quality.\
plot_multi_sds_colour_rendering_indexes_bars` definition unit tests methods.
    """

    def test_plot_multi_sds_colour_rendering_indexes_bars(self):
        """
        Test :func:`colour.plotting.quality.\
plot_multi_sds_colour_rendering_indexes_bars` definition.
        """

        figure, axes = plot_multi_sds_colour_rendering_indexes_bars(
            [SDS_ILLUMINANTS["FL2"], SDS_LIGHT_SOURCES["Kinoton 75P"]]
        )

        self.assertIsInstance(figure, Figure)
        self.assertIsInstance(axes, Axes)


class TestPlotSingleSdColourQualityScaleBars(unittest.TestCase):
    """
    Define :func:`colour.plotting.quality.\
plot_single_sd_colour_quality_scale_bars` definition unit tests methods.
    """

    def test_plot_single_sd_colour_quality_scale_bars(self):
        """
        Test :func:`colour.plotting.quality.\
plot_single_sd_colour_quality_scale_bars` definition.
        """

        figure, axes = plot_single_sd_colour_quality_scale_bars(
            SDS_ILLUMINANTS["FL2"]
        )

        self.assertIsInstance(figure, Figure)
        self.assertIsInstance(axes, Axes)


class TestPlotMultiSdsColourQualityScalesBars(unittest.TestCase):
    """
    Define :func:`colour.plotting.quality.\
plot_multi_sds_colour_quality_scales_bars` definition unit tests methods.
    """

    def test_plot_multi_sds_colour_quality_scales_bars(self):
        """
        Test :func:`colour.plotting.quality.\
plot_multi_sds_colour_quality_scales_bars` definition.
        """

        figure, axes = plot_multi_sds_colour_quality_scales_bars(
            [SDS_ILLUMINANTS["FL2"], SDS_LIGHT_SOURCES["Kinoton 75P"]]
        )

        self.assertIsInstance(figure, Figure)
        self.assertIsInstance(axes, Axes)


if __name__ == "__main__":
    unittest.main()



"""
Hunt Colour Appearance Model
============================

Defines the *Hunt* colour appearance model objects:

-   :class:`colour.appearance.InductionFactors_Hunt`
-   :attr:`colour.VIEWING_CONDITIONS_HUNT`
-   :class:`colour.CAM_Specification_Hunt`
-   :func:`colour.XYZ_to_Hunt`

References
----------
-   :cite:`Fairchild2013u` : Fairchild, M. D. (2013). The Hunt Model. In Color
    Appearance Models (3rd ed., pp. 5094-5556). Wiley. ISBN:B00DAYO8E2
-   :cite:`Hunt2004b` : Hunt, R. W. G. (2004). The Reproduction of Colour (6th
    ed.). John Wiley & Sons, Ltd. doi:10.1002/0470024275
"""

from __future__ import annotations

import numpy as np
from collections import namedtuple
from dataclasses import dataclass, field

from colour.algebra import spow, vector_dot
from colour.hints import (
    ArrayLike,
    Boolean,
    Dict,
    FloatingOrArrayLike,
    FloatingOrNDArray,
    NDArray,
    Optional,
    cast,
)
from colour.utilities import (
    CaseInsensitiveMapping,
    MixinDataclassArithmetic,
    as_float,
    as_float_array,
    from_range_degrees,
    ones,
    to_domain_100,
    tsplit,
    tstack,
    usage_warning,
    zeros,
)

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "InductionFactors_Hunt",
    "VIEWING_CONDITIONS_HUNT",
    "HUE_DATA_FOR_HUE_QUADRATURE",
    "MATRIX_XYZ_TO_HPE",
    "MATRIX_HPE_TO_XYZ",
    "CAM_ReferenceSpecification_Hunt",
    "CAM_Specification_Hunt",
    "XYZ_to_Hunt",
    "luminance_level_adaptation_factor",
    "illuminant_scotopic_luminance",
    "XYZ_to_rgb",
    "f_n",
    "chromatic_adaptation",
    "adjusted_reference_white_signals",
    "achromatic_post_adaptation_signal",
    "colour_difference_signals",
    "hue_angle",
    "eccentricity_factor",
    "low_luminance_tritanopia_factor",
    "yellowness_blueness_response",
    "redness_greenness_response",
    "overall_chromatic_response",
    "saturation_correlate",
    "achromatic_signal",
    "brightness_correlate",
    "lightness_correlate",
    "chroma_correlate",
    "colourfulness_correlate",
]


class InductionFactors_Hunt(
    namedtuple("InductionFactors_Hunt", ("N_c", "N_b", "N_cb", "N_bb"))
):
    """
    *Hunt* colour appearance model induction factors.

    Parameters
    ----------
    N_c
        Chromatic surround induction factor :math:`N_c`.
    N_b
        *Brightness* surround induction factor :math:`N_b`.
    N_cb
        Chromatic background induction factor :math:`N_{cb}`, approximated
        using tristimulus values :math:`Y_w` and :math:`Y_b` of
        respectively the reference white and the background if not specified.
    N_bb
        *Brightness* background induction factor :math:`N_{bb}`, approximated
        using tristimulus values :math:`Y_w` and :math:`Y_b` of
        respectively the reference white and the background if not specified.

    References
    ----------
    :cite:`Fairchild2013u`, :cite:`Hunt2004b`
    """

    def __new__(cls, N_c, N_b, N_cb=None, N_bb=None):
        """
        Return a new instance of the
        :class:`colour.appearance.InductionFactors_Hunt` class.
        """

        return super().__new__(cls, N_c, N_b, N_cb, N_bb)


VIEWING_CONDITIONS_HUNT: CaseInsensitiveMapping = CaseInsensitiveMapping(
    {
        "Small Areas, Uniform Background & Surrounds": InductionFactors_Hunt(
            1, 300
        ),
        "Normal Scenes": InductionFactors_Hunt(1, 75),
        "Television & CRT, Dim Surrounds": InductionFactors_Hunt(1, 25),
        "Large Transparencies On Light Boxes": InductionFactors_Hunt(0.7, 25),
        "Projected Transparencies, Dark Surrounds": InductionFactors_Hunt(
            0.7, 10
        ),
    }
)
VIEWING_CONDITIONS_HUNT.__doc__ = """
Reference *Hunt* colour appearance model viewing conditions.

References
----------
:cite:`Fairchild2013u`, :cite:`Hunt2004b`

Aliases:

-   'small_uniform': 'Small Areas, Uniform Background & Surrounds'
-   'normal': 'Normal Scenes'
-   'tv_dim': 'Television & CRT, Dim Surrounds'
-   'light_boxes': 'Large Transparencies On Light Boxes'
-   'projected_dark': 'Projected Transparencies, Dark Surrounds'

"""
VIEWING_CONDITIONS_HUNT["small_uniform"] = VIEWING_CONDITIONS_HUNT[
    "Small Areas, Uniform Background & Surrounds"
]
VIEWING_CONDITIONS_HUNT["normal"] = VIEWING_CONDITIONS_HUNT["Normal Scenes"]
VIEWING_CONDITIONS_HUNT["tv_dim"] = VIEWING_CONDITIONS_HUNT[
    "Television & CRT, Dim Surrounds"
]
VIEWING_CONDITIONS_HUNT["light_boxes"] = VIEWING_CONDITIONS_HUNT[
    "Large Transparencies On Light Boxes"
]
VIEWING_CONDITIONS_HUNT["projected_dark"] = VIEWING_CONDITIONS_HUNT[
    "Projected Transparencies, Dark Surrounds"
]

HUE_DATA_FOR_HUE_QUADRATURE: Dict = {
    "h_s": np.array([20.14, 90.00, 164.25, 237.53]),
    "e_s": np.array([0.8, 0.7, 1.0, 1.2]),
}

MATRIX_XYZ_TO_HPE: NDArray = np.array(
    [
        [0.38971, 0.68898, -0.07868],
        [-0.22981, 1.18340, 0.04641],
        [0.00000, 0.00000, 1.00000],
    ]
)
"""
*Hunt* colour appearance model *CIE XYZ* tristimulus values to
*Hunt-Pointer-Estevez* :math:`\\rho\\gamma\\beta` colourspace matrix.
"""

MATRIX_HPE_TO_XYZ: NDArray = np.linalg.inv(MATRIX_XYZ_TO_HPE)
"""
*Hunt* colour appearance model *Hunt-Pointer-Estevez*
:math:`\\rho\\gamma\\beta` colourspace to *CIE XYZ* tristimulus values matrix.
"""


@dataclass
class CAM_ReferenceSpecification_Hunt(MixinDataclassArithmetic):
    """
    Define the *Hunt* colour appearance model reference specification.

    This specification has field names consistent with *Fairchild (2013)*
    reference.

    Parameters
    ----------
    J
        Correlate of *Lightness* :math:`J`.
    C_94
        Correlate of *chroma* :math:`C_94`.
    h_S
        *Hue* angle :math:`h_S` in degrees.
    s
        Correlate of *saturation* :math:`s`.
    Q
        Correlate of *brightness* :math:`Q`.
    M_94
        Correlate of *colourfulness* :math:`M_94`.
    H
        *Hue* :math:`h` quadrature :math:`H`.
    H_C
        *Hue* :math:`h` composition :math:`H_C`.

    References
    ----------
    :cite:`Fairchild2013u`, :cite:`Hunt2004b`
    """

    J: Optional[FloatingOrNDArray] = field(default_factory=lambda: None)
    C_94: Optional[FloatingOrNDArray] = field(default_factory=lambda: None)
    h_S: Optional[FloatingOrNDArray] = field(default_factory=lambda: None)
    s: Optional[FloatingOrNDArray] = field(default_factory=lambda: None)
    Q: Optional[FloatingOrNDArray] = field(default_factory=lambda: None)
    M_94: Optional[FloatingOrNDArray] = field(default_factory=lambda: None)
    H: Optional[FloatingOrNDArray] = field(default_factory=lambda: None)
    H_C: Optional[FloatingOrNDArray] = field(default_factory=lambda: None)


@dataclass
class CAM_Specification_Hunt(MixinDataclassArithmetic):
    """
    Define the *Hunt* colour appearance model specification.

    This specification has field names consistent with the remaining colour
    appearance models in :mod:`colour.appearance` but diverge from
    *Fairchild (2013)* reference.

    Parameters
    ----------
    J
        Correlate of *Lightness* :math:`J`.
    C
        Correlate of *chroma* :math:`C_94`.
    h
        *Hue* angle :math:`h_S` in degrees.
    s
        Correlate of *saturation* :math:`s`.
    Q
        Correlate of *brightness* :math:`Q`.
    M
        Correlate of *colourfulness* :math:`M_94`.
    H
        *Hue* :math:`h` quadrature :math:`H`.
    HC
        *Hue* :math:`h` composition :math:`H_C`.

    Notes
    -----
    -   This specification is the one used in the current model implementation.

    References
    ----------
    :cite:`Fairchild2013u`, :cite:`Hunt2004b`
    """

    J: Optional[FloatingOrNDArray] = field(default_factory=lambda: None)
    C: Optional[FloatingOrNDArray] = field(default_factory=lambda: None)
    h: Optional[FloatingOrNDArray] = field(default_factory=lambda: None)
    s: Optional[FloatingOrNDArray] = field(default_factory=lambda: None)
    Q: Optional[FloatingOrNDArray] = field(default_factory=lambda: None)
    M: Optional[FloatingOrNDArray] = field(default_factory=lambda: None)
    H: Optional[FloatingOrNDArray] = field(default_factory=lambda: None)
    HC: Optional[FloatingOrNDArray] = field(default_factory=lambda: None)


def XYZ_to_Hunt(
    XYZ: ArrayLike,
    XYZ_w: ArrayLike,
    XYZ_b: ArrayLike,
    L_A: FloatingOrArrayLike,
    surround: InductionFactors_Hunt = VIEWING_CONDITIONS_HUNT["Normal Scenes"],
    L_AS: Optional[FloatingOrArrayLike] = None,
    CCT_w: Optional[FloatingOrArrayLike] = None,
    XYZ_p: Optional[ArrayLike] = None,
    p: Optional[FloatingOrArrayLike] = None,
    S: Optional[FloatingOrArrayLike] = None,
    S_w: Optional[FloatingOrArrayLike] = None,
    helson_judd_effect: Boolean = False,
    discount_illuminant: Boolean = True,
) -> CAM_Specification_Hunt:
    """
    Compute the *Hunt* colour appearance model correlates.

    Parameters
    ----------
    XYZ
        *CIE XYZ* tristimulus values of test sample / stimulus.
    XYZ_w
        *CIE XYZ* tristimulus values of reference white.
    XYZ_b
        *CIE XYZ* tristimulus values of background.
    L_A
        Adapting field *luminance* :math:`L_A` in :math:`cd/m^2`.
    surround
         Surround viewing conditions induction factors.
    L_AS
        Scotopic luminance :math:`L_{AS}` of the illuminant, approximated if
        not specified.
    CCT_w
        Correlated color temperature :math:`T_{cp}`: of the illuminant, needed
        to approximate :math:`L_{AS}`.
    XYZ_p
        *CIE XYZ* tristimulus values of proximal field, assumed to be equal to
        background if not specified.
    p
        Simultaneous contrast / assimilation factor :math:`p` with value
        normalised to domain [-1, 0] when simultaneous contrast occurs and
        normalised to domain [0, 1] when assimilation occurs.
    S
        Scotopic response :math:`S` to the stimulus, approximated using
        tristimulus values :math:`Y` of the stimulus if not specified.
    S_w
        Scotopic response :math:`S_w` for the reference white, approximated
        using the tristimulus values :math:`Y_w` of the reference white if not
        specified.
    helson_judd_effect
        Truth value indicating whether the *Helson-Judd* effect should be
        accounted for.
    discount_illuminant
       Truth value indicating if the illuminant should be discounted.

    Returns
    -------
    :class:`colour.CAM_Specification_Hunt`
        *Hunt* colour appearance model specification.

    Raises
    ------
    ValueError
        If an illegal argument combination is specified.

    Notes
    -----
    +------------+-----------------------+---------------+
    | **Domain** | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``XYZ``    | [0, 100]              | [0, 1]        |
    +------------+-----------------------+---------------+
    | ``XYZ_w``  | [0, 100]              | [0, 1]        |
    +------------+-----------------------+---------------+
    | ``XYZ_b``  | [0, 100]              | [0, 1]        |
    +------------+-----------------------+---------------+
    | ``XYZ_p``  | [0, 100]              | [0, 1]        |
    +------------+-----------------------+---------------+

    +------------------------------+-----------------------+---------------+
    | **Range**                    | **Scale - Reference** | **Scale - 1** |
    +==============================+=======================+===============+
    | ``CAM_Specification_Hunt.h`` | [0, 360]              | [0, 1]        |
    +------------------------------+-----------------------+---------------+

    References
    ----------
    :cite:`Fairchild2013u`, :cite:`Hunt2004b`

    Examples
    --------
    >>> XYZ = np.array([19.01, 20.00, 21.78])
    >>> XYZ_w = np.array([95.05, 100.00, 108.88])
    >>> XYZ_b = np.array([95.05, 100.00, 108.88])
    >>> L_A = 318.31
    >>> surround = VIEWING_CONDITIONS_HUNT['Normal Scenes']
    >>> CCT_w = 6504
    >>> XYZ_to_Hunt(XYZ, XYZ_w, XYZ_b, L_A, surround, CCT_w=CCT_w)
    ... # doctest: +ELLIPSIS
    CAM_Specification_Hunt(J=30.0462678..., C=0.1210508..., h=269.2737594..., \
s=0.0199093..., Q=22.2097654..., M=0.1238964..., H=None, HC=None)
    """

    XYZ = to_domain_100(XYZ)
    XYZ_w = to_domain_100(XYZ_w)
    XYZ_b = to_domain_100(XYZ_b)
    _X, Y, _Z = tsplit(XYZ)
    _X_w, Y_w, _Z_w = tsplit(XYZ_w)
    X_b, Y_b, _Z_b = tsplit(XYZ_b)

    # Arguments handling.
    if XYZ_p is not None:
        X_p, Y_p, Z_p = tsplit(to_domain_100(XYZ_p))
    else:
        X_p = X_b
        Y_p = Y_b
        Z_p = Y_b
        usage_warning(
            'Unspecified proximal field "XYZ_p" argument, using '
            'background "XYZ_b" as approximation!'
        )

    if surround.N_cb is None:
        N_cb = 0.725 * spow(Y_w / Y_b, 0.2)
        usage_warning(
            f'Unspecified "N_cb" argument, using approximation: "{N_cb}"'
        )
    if surround.N_bb is None:
        N_bb = 0.725 * spow(Y_w / Y_b, 0.2)
        usage_warning(
            f'Unspecified "N_bb" argument, using approximation: "{N_bb}"'
        )

    if L_AS is None and CCT_w is None:
        raise ValueError(
            'Either the scotopic luminance "L_AS" of the '
            "illuminant or its correlated colour temperature "
            '"CCT_w" must be specified!'
        )
    elif L_AS is None and CCT_w is not None:
        L_AS = illuminant_scotopic_luminance(L_A, CCT_w)
        usage_warning(
            f'Unspecified "L_AS" argument, using approximation from "CCT": '
            f'"{L_AS}"'
        )

    if (S is None and S_w is not None) or (S is not None and S_w is None):
        raise ValueError(
            'Either both stimulus scotopic response "S" and '
            'reference white scotopic response "S_w" arguments '
            "need to be specified or none of them!"
        )
    elif S is None and S_w is None:
        S_p = Y
        S_w_p = Y_w
        usage_warning(
            f'Unspecified stimulus scotopic response "S" and reference white '
            f'scotopic response "S_w" arguments, using approximation: '
            f'"{S}", "{S_w}"'
        )

    if p is None:
        usage_warning(
            'Unspecified simultaneous contrast / assimilation "p" '
            "argument, model will not account for simultaneous chromatic "
            "contrast!"
        )

    XYZ_p = tstack([X_p, Y_p, Z_p])

    # Computing luminance level adaptation factor :math:`F_L`.
    F_L = luminance_level_adaptation_factor(L_A)

    # Computing test sample chromatic adaptation.
    rgb_a = chromatic_adaptation(
        XYZ,
        XYZ_w,
        XYZ_b,
        L_A,
        F_L,
        XYZ_p,
        p,
        helson_judd_effect,
        discount_illuminant,
    )

    # Computing reference white chromatic adaptation.
    rgb_aw = chromatic_adaptation(
        XYZ_w,
        XYZ_w,
        XYZ_b,
        L_A,
        F_L,
        XYZ_p,
        p,
        helson_judd_effect,
        discount_illuminant,
    )

    # Computing opponent colour dimensions.
    # Computing achromatic post adaptation signals.
    A_a = achromatic_post_adaptation_signal(rgb_a)
    A_aw = achromatic_post_adaptation_signal(rgb_aw)

    # Computing colour difference signals.
    C = colour_difference_signals(rgb_a)
    C_w = colour_difference_signals(rgb_aw)

    # -------------------------------------------------------------------------
    # Computing the *hue* angle :math:`h_s`.
    # -------------------------------------------------------------------------
    h = hue_angle(C)
    # hue_w = hue_angle(C_w)
    # TODO: Implement hue quadrature & composition computation.

    # -------------------------------------------------------------------------
    # Computing the correlate of *saturation* :math:`s`.
    # -------------------------------------------------------------------------
    # Computing eccentricity factors.
    e_s = eccentricity_factor(h)

    # Computing low luminance tritanopia factor :math:`F_t`.
    F_t = low_luminance_tritanopia_factor(L_A)

    M_yb = yellowness_blueness_response(C, e_s, surround.N_c, N_cb, F_t)
    M_rg = redness_greenness_response(C, e_s, surround.N_c, N_cb)
    M_yb_w = yellowness_blueness_response(C_w, e_s, surround.N_c, N_cb, F_t)
    M_rg_w = redness_greenness_response(C_w, e_s, surround.N_c, N_cb)

    # Computing overall chromatic response.
    M = overall_chromatic_response(M_yb, M_rg)
    M_w = overall_chromatic_response(M_yb_w, M_rg_w)

    s = saturation_correlate(M, rgb_a)

    # -------------------------------------------------------------------------
    # Computing the correlate of *brightness* :math:`Q`.
    # -------------------------------------------------------------------------
    # Computing achromatic signal :math:`A`.
    A = achromatic_signal(cast(FloatingOrNDArray, L_AS), S_p, S_w_p, N_bb, A_a)
    A_w = achromatic_signal(
        cast(FloatingOrNDArray, L_AS), S_w_p, S_w_p, N_bb, A_aw
    )

    Q = brightness_correlate(A, A_w, M, surround.N_b)
    brightness_w = brightness_correlate(A_w, A_w, M_w, surround.N_b)
    # TODO: Implement whiteness-blackness :math:`Q_{wb}` computation.

    # -------------------------------------------------------------------------
    # Computing the correlate of *Lightness* :math:`J`.
    # -------------------------------------------------------------------------
    J = lightness_correlate(Y_b, Y_w, Q, brightness_w)

    # -------------------------------------------------------------------------
    # Computing the correlate of *chroma* :math:`C_{94}`.
    # -------------------------------------------------------------------------
    C_94 = chroma_correlate(s, Y_b, Y_w, Q, brightness_w)

    # -------------------------------------------------------------------------
    # Computing the correlate of *colourfulness* :math:`M_{94}`.
    # -------------------------------------------------------------------------
    M_94 = colourfulness_correlate(F_L, C_94)

    return CAM_Specification_Hunt(
        J,
        C_94,
        as_float(from_range_degrees(h)),
        s,
        Q,
        M_94,
        None,
        None,
    )


def luminance_level_adaptation_factor(
    L_A: FloatingOrArrayLike,
) -> FloatingOrNDArray:
    """
    Return the *luminance* level adaptation factor :math:`F_L`.

    Parameters
    ----------
    L_A
        Adapting field *luminance* :math:`L_A` in :math:`cd/m^2`.

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        *Luminance* level adaptation factor :math:`F_L`

    Examples
    --------
    >>> luminance_level_adaptation_factor(318.31)  # doctest: +ELLIPSIS
    1.1675444...
    """

    L_A = as_float_array(L_A)

    k = 1 / (5 * L_A + 1)
    k4 = k**4
    F_L = 0.2 * k4 * (5 * L_A) + 0.1 * (1 - k4) ** 2 * spow(5 * L_A, 1 / 3)

    return as_float(F_L)


def illuminant_scotopic_luminance(
    L_A: FloatingOrArrayLike, CCT: FloatingOrArrayLike
) -> FloatingOrNDArray:
    """
    Return the approximate scotopic luminance :math:`L_{AS}` of the
    illuminant.

    Parameters
    ----------
    L_A
        Adapting field *luminance* :math:`L_A` in :math:`cd/m^2`.
    CCT
        Correlated color temperature :math:`T_{cp}` of the illuminant.

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        Approximate scotopic luminance :math:`L_{AS}`.

    Examples
    --------
    >>> illuminant_scotopic_luminance(318.31, 6504.0)  # doctest: +ELLIPSIS
    769.9376286...
    """

    L_A = as_float_array(L_A)
    CCT = as_float_array(CCT)

    CCT = 2.26 * L_A * spow((CCT / 4000) - 0.4, 1 / 3)

    return as_float(CCT)


def XYZ_to_rgb(XYZ: ArrayLike) -> NDArray:
    """
    Convert from *CIE XYZ* tristimulus values to *Hunt-Pointer-Estevez*
    :math:`\\rho\\gamma\\beta` colourspace.

    Parameters
    ----------
    XYZ
        *CIE XYZ* tristimulus values.

    Returns
    -------
    :class:`numpy.ndarray`
        *Hunt-Pointer-Estevez* :math:`\\rho\\gamma\\beta` colourspace.

    Examples
    --------
    >>> XYZ = np.array([19.01, 20.00, 21.78])
    >>> XYZ_to_rgb(XYZ)  # doctest: +ELLIPSIS
    array([ 19.4743367...,  20.3101217...,  21.78     ])
    """

    return vector_dot(MATRIX_XYZ_TO_HPE, XYZ)


def f_n(x: ArrayLike) -> NDArray:
    """
    Define the nonlinear response function of the *Hunt* colour appearance
    model used to model the nonlinear behaviour of various visual responses.

    Parameters
    ----------
    x
        Visual response variable :math:`x`.

    Returns
    -------
    :class:`numpy.ndarray`
        Modeled visual response variable :math:`x`.


    Examples
    --------
    >>> x = np.array([0.23350512, 0.23351103, 0.23355179])
    >>> f_n(x)  # doctest: +ELLIPSIS
    array([ 5.8968592...,  5.8969521...,  5.8975927...])
    """

    x_p = spow(x, 0.73)
    x_m = 40 * (x_p / (x_p + 2))

    return as_float_array(x_m)


def chromatic_adaptation(
    XYZ: ArrayLike,
    XYZ_w: ArrayLike,
    XYZ_b: ArrayLike,
    L_A,
    F_L,
    XYZ_p: ArrayLike = None,
    p: Optional[FloatingOrArrayLike] = None,
    helson_judd_effect: Boolean = False,
    discount_illuminant: Boolean = True,
) -> NDArray:
    """
    Apply chromatic adaptation to given *CIE XYZ* tristimulus values.

    Parameters
    ----------
    XYZ
        *CIE XYZ* tristimulus values of test sample.
    XYZ_b
        *CIE XYZ* tristimulus values of background.
    XYZ_w
        *CIE XYZ* tristimulus values of reference white.
    L_A
        Adapting field *luminance* :math:`L_A` in :math:`cd/m^2`.
    F_L
        Luminance adaptation factor :math:`F_L`.
    XYZ_p
        *CIE XYZ* tristimulus values of proximal field, assumed to be equal to
        background if not specified.
    p
        Simultaneous contrast / assimilation factor :math:`p` with value
        normalised to  domain [-1, 0] when simultaneous contrast occurs and
        normalised to domain [0, 1] when assimilation occurs.
    helson_judd_effect
        Truth value indicating whether the *Helson-Judd* effect should be
        accounted for.
    discount_illuminant
       Truth value indicating if the illuminant should be discounted.

    Returns
    -------
    :class:`numpy.ndarray`
        Adapted *CIE XYZ* tristimulus values.

    Examples
    --------
    >>> XYZ = np.array([19.01, 20.00, 21.78])
    >>> XYZ_b = np.array([95.05, 100.00, 108.88])
    >>> XYZ_w = np.array([95.05, 100.00, 108.88])
    >>> L_A = 318.31
    >>> F_L = 1.16754446415
    >>> chromatic_adaptation(XYZ, XYZ_w, XYZ_b, L_A, F_L)  # doctest: +ELLIPSIS
    array([ 6.8959454...,  6.8959991...,  6.8965708...])

    # Coverage Doctests

    >>> chromatic_adaptation(XYZ, XYZ_w, XYZ_b, L_A, F_L,
    ...                      discount_illuminant=False)  # doctest: +ELLIPSIS
    array([ 6.8525880...,  6.8874417...,  6.9461478...])
    >>> chromatic_adaptation(XYZ, XYZ_w, XYZ_b, L_A, F_L,
    ...                      helson_judd_effect=True)  # doctest: +ELLIPSIS
    array([ 6.8959454...,  6.8959991...,  6.8965708...])
    >>> chromatic_adaptation(XYZ, XYZ_w, XYZ_b, L_A, F_L,
    ...                      XYZ_p=XYZ_b, p=0.5)  # doctest: +ELLIPSIS
    array([ 9.2069020...,  9.2070219...,  9.2078373...])
    """

    XYZ_w = as_float_array(XYZ_w)
    XYZ_b = as_float_array(XYZ_b)
    L_A = as_float_array(L_A)
    F_L = as_float_array(F_L)

    rgb = XYZ_to_rgb(XYZ)
    rgb_w = XYZ_to_rgb(XYZ_w)
    Y_w = XYZ_w[..., 1]
    Y_b = XYZ_b[..., 1]

    h_rgb = 3 * rgb_w / np.sum(rgb_w, axis=-1)[..., np.newaxis]

    # Computing chromatic adaptation factors.
    if not discount_illuminant:
        L_A_p = spow(L_A, 1 / 3)
        F_rgb = (1 + L_A_p + h_rgb) / (1 + L_A_p + (1 / h_rgb))
    else:
        F_rgb = ones(h_rgb.shape)

    # Computing Helson-Judd effect parameters.
    if helson_judd_effect:
        D_rgb = f_n((Y_b / Y_w) * F_L * F_rgb[..., 1]) - f_n(
            (Y_b / Y_w) * F_L * F_rgb
        )
    else:
        D_rgb = zeros(F_rgb.shape)

    # Computing cone bleach factors.
    B_rgb = (10**7) / ((10**7) + 5 * L_A[..., np.newaxis] * (rgb_w / 100))

    # Computing adjusted reference white signals.
    if XYZ_p is not None and p is not None:
        rgb_p = XYZ_to_rgb(XYZ_p)
        rgb_w = adjusted_reference_white_signals(rgb_p, B_rgb, rgb_w, p)

    # Computing adapted cone responses.
    rgb_a = 1 + B_rgb * (
        f_n(F_L[..., np.newaxis] * F_rgb * rgb / rgb_w) + D_rgb
    )

    return rgb_a


def adjusted_reference_white_signals(
    rgb_p: ArrayLike,
    rgb_b: ArrayLike,
    rgb_w: ArrayLike,
    p: FloatingOrArrayLike,
) -> NDArray:
    """
    Adjust the white point for simultaneous chromatic contrast.

    Parameters
    ----------
    rgb_p
        Cone signals *Hunt-Pointer-Estevez* :math:`\\rho\\gamma\\beta`
        colourspace array of the proximal field.
    rgb_b
        Cone signals *Hunt-Pointer-Estevez* :math:`\\rho\\gamma\\beta`
        colourspace array of the background.
    rgb_w
        Cone signals array *Hunt-Pointer-Estevez* :math:`\\rho\\gamma\\beta`
        colourspace array of the reference white.
    p
        Simultaneous contrast / assimilation factor :math:`p` with value
        normalised to domain [-1, 0] when simultaneous contrast occurs and
        normalised to domain [0, 1] when assimilation occurs.

    Returns
    -------
    :class:`numpy.ndarray`
        Adjusted cone signals *Hunt-Pointer-Estevez* :math:`\\rho\\gamma\\beta`
        colourspace array of the reference white.

    Examples
    --------
    >>> rgb_p = np.array([98.07193550, 101.13755950, 100.00000000])
    >>> rgb_b = np.array([0.99984505, 0.99983840, 0.99982674])
    >>> rgb_w = np.array([97.37325710, 101.54968030, 108.88000000])
    >>> p = 0.1
    >>> adjusted_reference_white_signals(rgb_p, rgb_b, rgb_w, p)
    ... # doctest: +ELLIPSIS
    array([ 88.0792742...,  91.8569553...,  98.4876543...])
    """

    rgb_p = as_float_array(rgb_p)
    rgb_b = as_float_array(rgb_b)
    rgb_w = as_float_array(rgb_w)
    p = as_float_array(p)

    p_rgb = rgb_p / rgb_b
    rgb_w = (
        rgb_w
        * (spow((1 - p) * p_rgb + (1 + p) / p_rgb, 0.5))
        / (spow((1 + p) * p_rgb + (1 - p) / p_rgb, 0.5))
    )

    return rgb_w


def achromatic_post_adaptation_signal(rgb: ArrayLike) -> NDArray:
    """
    Return the achromatic post adaptation signal :math:`A` from given
    *Hunt-Pointer-Estevez* :math:`\\rho\\gamma\\beta` colourspace array.

    Parameters
    ----------
    rgb
        *Hunt-Pointer-Estevez* :math:`\\rho\\gamma\\beta` colourspace array.

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        Achromatic post adaptation signal :math:`A`.

    Examples
    --------
    >>> rgb = np.array([6.89594549, 6.89599915, 6.89657085])
    >>> achromatic_post_adaptation_signal(rgb)  # doctest: +ELLIPSIS
    18.9827186...
    """

    r, g, b = tsplit(rgb)

    A = 2 * r + g + (1 / 20) * b - 3.05 + 1

    return A


def colour_difference_signals(rgb: ArrayLike) -> NDArray:
    """
    Return the colour difference signals :math:`C_1`, :math:`C_2` and
    :math:`C_3` from given *Hunt-Pointer-Estevez* :math:`\\rho\\gamma\\beta`
    colourspace array.

    Parameters
    ----------
    rgb
        *Hunt-Pointer-Estevez* :math:`\\rho\\gamma\\beta` colourspace array.

    Returns
    -------
    :class:`numpy.ndarray`
        Colour difference signals :math:`C_1`, :math:`C_2` and :math:`C_3`.

    Examples
    --------
    >>> rgb = np.array([6.89594549, 6.89599915, 6.89657085])
    >>> colour_difference_signals(rgb)  # doctest: +ELLIPSIS
    array([ -5.3660000...e-05,  -5.7170000...e-04,   6.2536000...e-04])
    """

    r, g, b = tsplit(rgb)

    C_1 = r - g
    C_2 = g - b
    C_3 = b - r

    C = tstack([C_1, C_2, C_3])

    return C


def hue_angle(C: FloatingOrArrayLike) -> FloatingOrNDArray:
    """
    Return the *hue* angle :math:`h` in degrees from given colour difference
    signals :math:`C`.

    Parameters
    ----------
    C
        Colour difference signals :math:`C`.

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        *Hue* angle :math:`h` in degrees.

    Examples
    --------
    >>> C = np.array([
    ...     -5.365865581996587e-05,
    ...     -0.000571699383647,
    ...     0.000625358039467
    ... ])
    >>> hue_angle(C)  # doctest: +ELLIPSIS
    269.2737594...
    """

    C_1, C_2, C_3 = tsplit(C)

    hue = (
        180 * np.arctan2(0.5 * (C_2 - C_3) / 4.5, C_1 - (C_2 / 11)) / np.pi
    ) % 360
    return as_float(hue)


def eccentricity_factor(hue: FloatingOrArrayLike) -> FloatingOrNDArray:
    """
    Return eccentricity factor :math:`e_s` from given hue angle :math:`h`
    in degrees.

    Parameters
    ----------
    hue
        Hue angle :math:`h` in degrees.

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        Eccentricity factor :math:`e_s`.

    Examples
    --------
    >>> eccentricity_factor(269.273759)  # doctest: +ELLIPSIS
    1.1108365...
    """

    hue = as_float_array(hue)

    h_s = HUE_DATA_FOR_HUE_QUADRATURE["h_s"]
    e_s = HUE_DATA_FOR_HUE_QUADRATURE["e_s"]

    x = np.interp(hue, h_s, e_s)
    x = np.where(hue < 20.14, 0.856 - (hue / 20.14) * 0.056, x)
    x = np.where(hue > 237.53, 0.856 + 0.344 * (360 - hue) / (360 - 237.53), x)

    return as_float(x)


def low_luminance_tritanopia_factor(
    L_A: FloatingOrArrayLike,
) -> FloatingOrNDArray:
    """
    Return the low luminance tritanopia factor :math:`F_t` from given adapting
    field *luminance* :math:`L_A` in :math:`cd/m^2`.

    Parameters
    ----------
    L_A
        Adapting field *luminance* :math:`L_A` in :math:`cd/m^2`.

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        Low luminance tritanopia factor :math:`F_t`.

    Examples
    --------
    >>> low_luminance_tritanopia_factor(318.31)  # doctest: +ELLIPSIS
    0.9996859...
    """

    L_A = as_float_array(L_A)

    F_t = L_A / (L_A + 0.1)

    return as_float(F_t)


def yellowness_blueness_response(
    C: FloatingOrArrayLike,
    e_s: FloatingOrArrayLike,
    N_c: FloatingOrArrayLike,
    N_cb: FloatingOrArrayLike,
    F_t: FloatingOrArrayLike,
) -> FloatingOrNDArray:
    """
    Return the yellowness / blueness response :math:`M_{yb}`.

    Parameters
    ----------
    C
        Colour difference signals :math:`C`.
    e_s
        Eccentricity factor :math:`e_s`.
    N_c
         Chromatic surround induction factor :math:`N_c`.
    N_cb
         Chromatic background induction factor :math:`N_{cb}`.
    F_t
        Low luminance tritanopia factor :math:`F_t`.

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        Yellowness / blueness response :math:`M_{yb}`.

    Examples
    --------
    >>> C = np.array([
    ...     -5.365865581996587e-05,
    ...     -0.000571699383647,
    ...     0.000625358039467
    ... ])
    >>> e_s = 1.110836504862630
    >>> N_c = 1.0
    >>> N_cb = 0.725000000000000
    >>> F_t = 0.99968593951195
    >>> yellowness_blueness_response(C, e_s, N_c, N_cb, F_t)
    ... # doctest: +ELLIPSIS
    -0.0082372...
    """

    _C_1, C_2, C_3 = tsplit(C)
    e_s = as_float_array(e_s)
    N_c = as_float_array(N_c)
    N_cb = as_float_array(N_cb)
    F_t = as_float_array(F_t)

    M_yb = (
        100 * (0.5 * (C_2 - C_3) / 4.5) * (e_s * (10 / 13) * N_c * N_cb * F_t)
    )

    return as_float(M_yb)


def redness_greenness_response(
    C: FloatingOrArrayLike,
    e_s: FloatingOrArrayLike,
    N_c: FloatingOrArrayLike,
    N_cb: FloatingOrArrayLike,
) -> FloatingOrNDArray:
    """
    Return the redness / greenness response :math:`M_{yb}`.

    Parameters
    ----------
    C
        Colour difference signals :math:`C`.
    e_s
        Eccentricity factor :math:`e_s`.
    N_c
         Chromatic surround induction factor :math:`N_c`.
    N_cb
         Chromatic background induction factor :math:`N_{cb}`.

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        Redness / greenness response :math:`M_{rg}`.

    Examples
    --------
    >>> C = np.array([
    ...     -5.365865581996587e-05,
    ...     -0.000571699383647,
    ...     0.000625358039467
    ... ])
    >>> e_s = 1.110836504862630
    >>> N_c = 1.0
    >>> N_cb = 0.725000000000000
    >>> redness_greenness_response(C, e_s, N_c, N_cb)  # doctest: +ELLIPSIS
    -0.0001044...
    """

    C_1, C_2, _C_3 = tsplit(C)
    e_s = as_float_array(e_s)
    N_c = as_float_array(N_c)
    N_cb = as_float_array(N_cb)

    M_rg = 100 * (C_1 - (C_2 / 11)) * (e_s * (10 / 13) * N_c * N_cb)

    return as_float(M_rg)


def overall_chromatic_response(
    M_yb: FloatingOrArrayLike, M_rg: FloatingOrArrayLike
) -> FloatingOrNDArray:
    """
    Return the overall chromatic response :math:`M`.

    Parameters
    ----------
    M_yb
         Yellowness / blueness response :math:`M_{yb}`.
    M_rg
         Redness / greenness response :math:`M_{rg}`.

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        Overall chromatic response :math:`M`.

    Examples
    --------
    >>> M_yb = -0.008237223618825
    >>> M_rg = -0.000104447583276
    >>> overall_chromatic_response(M_yb, M_rg)  # doctest: +ELLIPSIS
    0.0082378...
    """

    M_yb = as_float_array(M_yb)
    M_rg = as_float_array(M_rg)

    M = spow((M_yb**2) + (M_rg**2), 0.5)

    return M


def saturation_correlate(
    M: FloatingOrArrayLike, rgb_a: ArrayLike
) -> FloatingOrNDArray:
    """
    Return the *saturation* correlate :math:`s`.

    Parameters
    ----------
    M
         Overall chromatic response :math:`M`.
    rgb_a
        Adapted *Hunt-Pointer-Estevez* :math:`\\rho\\gamma\\beta` colourspace
        array.

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        *Saturation* correlate :math:`s`.

    Examples
    --------
    >>> M = 0.008237885787274
    >>> rgb_a = np.array([6.89594549, 6.89599915, 6.89657085])
    >>> saturation_correlate(M, rgb_a)  # doctest: +ELLIPSIS
    0.0199093...
    """

    M = as_float_array(M)
    rgb_a = as_float_array(rgb_a)

    s = 50 * M / np.sum(rgb_a, axis=-1)

    return as_float(s)


def achromatic_signal(
    L_AS: FloatingOrArrayLike,
    S: FloatingOrArrayLike,
    S_w: FloatingOrArrayLike,
    N_bb: FloatingOrArrayLike,
    A_a: FloatingOrArrayLike,
) -> FloatingOrNDArray:
    """
    Return the achromatic signal :math:`A`.

    Parameters
    ----------
    L_AS
        Scotopic luminance :math:`L_{AS}` of the illuminant.
    S
        Scotopic response :math:`S` to the stimulus.
    S_w
        Scotopic response :math:`S_w` for the reference white.
    N_bb
        Brightness background induction factor :math:`N_{bb}`.
    A_a
        Achromatic post adaptation signal of the stimulus :math:`A_a`.

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        Achromatic signal :math:`A`.

    Examples
    --------
    >>> L_AS = 769.9376286541402
    >>> S = 20.0
    >>> S_w = 100.0
    >>> N_bb = 0.725000000000000
    >>> A_a = 18.982718664838487
    >>> achromatic_signal(L_AS, S, S_w, N_bb, A_a)  # doctest: +ELLIPSIS
    15.5068546...
    """

    L_AS = as_float_array(L_AS)
    S = as_float_array(S)
    S_w = as_float_array(S_w)
    N_bb = as_float_array(N_bb)
    A_a = as_float_array(A_a)

    j = 0.00001 / ((5 * L_AS / 2.26) + 0.00001)

    # Computing scotopic luminance level adaptation factor :math:`F_{LS}`.
    F_LS = 3800 * (j**2) * (5 * L_AS / 2.26)
    F_LS += 0.2 * (spow(1 - (j**2), 0.4)) * (spow(5 * L_AS / 2.26, 1 / 6))

    # Computing cone bleach factors :math:`B_S`.
    B_S = 0.5 / (1 + 0.3 * spow((5 * L_AS / 2.26) * (S / S_w), 0.3))
    B_S += 0.5 / (1 + 5 * (5 * L_AS / 2.26))

    # Computing adapted scotopic signal :math:`A_S`.
    A_S = (f_n(F_LS * S / S_w) * 3.05 * B_S) + 0.3

    # Computing achromatic signal :math:`A`.
    A = N_bb * (A_a - 1 + A_S - 0.3 + np.sqrt(1 + (0.3**2)))

    return as_float(A)


def brightness_correlate(
    A: FloatingOrArrayLike,
    A_w: FloatingOrArrayLike,
    M: FloatingOrArrayLike,
    N_b: FloatingOrArrayLike,
) -> FloatingOrNDArray:
    """
    Return the *brightness* correlate :math:`Q`.

    Parameters
    ----------
    A
         Achromatic signal :math:`A`.
    A_w
        Achromatic post adaptation signal of the reference white :math:`A_w`.
    M
        Overall chromatic response :math:`M`.
    N_b
         Brightness surround induction factor :math:`N_b`.

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        *Brightness* correlate :math:`Q`.

    Examples
    --------
    >>> A = 15.506854623621885
    >>> A_w = 35.718916676317086
    >>> M = 0.008237885787274
    >>> N_b = 75.0
    >>> brightness_correlate(A, A_w, M, N_b)  # doctest: +ELLIPSIS
    22.2097654...
    """

    A = as_float_array(A)
    A_w = as_float_array(A_w)
    M = as_float_array(M)
    N_b = as_float_array(N_b)

    N_1 = (spow(7 * A_w, 0.5)) / (5.33 * spow(N_b, 0.13))
    N_2 = (7 * A_w * spow(N_b, 0.362)) / 200

    Q = spow(7 * (A + (M / 100)), 0.6) * N_1 - N_2

    return Q


def lightness_correlate(
    Y_b: FloatingOrArrayLike,
    Y_w: FloatingOrArrayLike,
    Q: FloatingOrArrayLike,
    Q_w: FloatingOrArrayLike,
) -> FloatingOrNDArray:
    """
    Return the *Lightness* correlate :math:`J`.

    Parameters
    ----------
    Y_b
         Tristimulus values :math:`Y_b` the background.
    Y_w
         Tristimulus values :math:`Y_b` the reference white.
    Q
        *Brightness* correlate :math:`Q` of the stimulus.
    Q_w
        *Brightness* correlate :math:`Q` of the reference white.

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        *Lightness* correlate :math:`J`.

    Examples
    --------
    >>> Y_b = 100.0
    >>> Y_w = 100.0
    >>> Q = 22.209765491265024
    >>> Q_w = 40.518065821226081
    >>> lightness_correlate(Y_b, Y_w, Q, Q_w)  # doctest: +ELLIPSIS
    30.0462678...
    """

    Y_b = as_float_array(Y_b)
    Y_w = as_float_array(Y_w)
    Q = as_float_array(Q)
    Q_w = as_float_array(Q_w)

    Z = 1 + spow(Y_b / Y_w, 0.5)
    J = 100 * spow(Q / Q_w, Z)

    return J


def chroma_correlate(
    s: FloatingOrArrayLike,
    Y_b: FloatingOrArrayLike,
    Y_w: FloatingOrArrayLike,
    Q: FloatingOrArrayLike,
    Q_w: FloatingOrArrayLike,
) -> FloatingOrNDArray:
    """
    Return the *chroma* correlate :math:`C_94`.

    Parameters
    ----------
    s
        *Saturation* correlate :math:`s`.
    Y_b
         Tristimulus values :math:`Y_b` the background.
    Y_w
         Tristimulus values :math:`Y_b` the reference white.
    Q
        *Brightness* correlate :math:`Q` of the stimulus.
    Q_w
        *Brightness* correlate :math:`Q` of the reference white.

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        *Chroma* correlate :math:`C_94`.

    Examples
    --------
    >>> s = 0.0199093206929
    >>> Y_b = 100.0
    >>> Y_w = 100.0
    >>> Q = 22.209765491265024
    >>> Q_w = 40.518065821226081
    >>> chroma_correlate(s, Y_b, Y_w, Q, Q_w)  # doctest: +ELLIPSIS
    0.1210508...
    """

    s = as_float_array(s)
    Y_b = as_float_array(Y_b)
    Y_w = as_float_array(Y_w)
    Q = as_float_array(Q)
    Q_w = as_float_array(Q_w)

    C_94 = (
        2.44
        * spow(s, 0.69)
        * (spow(Q / Q_w, Y_b / Y_w))
        * (1.64 - spow(0.29, Y_b / Y_w))
    )

    return C_94


def colourfulness_correlate(
    F_L: FloatingOrArrayLike, C_94: FloatingOrArrayLike
) -> FloatingOrNDArray:
    """
    Return the *colourfulness* correlate :math:`M_94`.

    Parameters
    ----------
    F_L
        Luminance adaptation factor :math:`F_L`.
    C_94
        *Chroma* correlate :math:`C_94`.

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        *Colourfulness* correlate :math:`M_94`.

    Examples
    --------
    >>> F_L = 1.16754446414718
    >>> C_94 = 0.121050839936176
    >>> colourfulness_correlate(F_L, C_94)  # doctest: +ELLIPSIS
    0.1238964...
    """

    F_L = as_float_array(F_L)
    C_94 = as_float_array(C_94)

    M_94 = spow(F_L, 0.15) * C_94

    return M_94



"""
Extrapolation
=============

Defines the classes for extrapolating variables:

-   :class:`colour.Extrapolator`: 1-D function extrapolation.

References
----------
-   :cite:`Sastanina` : sastanin. (n.d.). How to make scipy.interpolate give an
    extrapolated result beyond the input range? Retrieved August 8, 2014, from
    http://stackoverflow.com/a/2745496/931625
-   :cite:`Westland2012i` : Westland, S., Ripamonti, C., & Cheung, V. (2012).
    Extrapolation Methods. In Computational Colour Science Using MATLAB (2nd
    ed., p. 38). ISBN:978-0-470-66569-5
"""

from __future__ import annotations

import numpy as np

from colour.algebra import NullInterpolator
from colour.constants import DEFAULT_FLOAT_DTYPE
from colour.hints import (
    DTypeNumber,
    FloatingOrArrayLike,
    FloatingOrNDArray,
    Literal,
    NDArray,
    Number,
    Optional,
    Type,
    TypeInterpolator,
    Union,
    cast,
)
from colour.utilities import (
    as_float,
    attest,
    is_numeric,
    is_string,
    optional,
    validate_method,
)

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "Extrapolator",
]


class Extrapolator:
    """
    Extrapolate the 1-D function of given interpolator.

    The :class:`colour.Extrapolator` class acts as a wrapper around a given
    *Colour* or *scipy* interpolator class instance with compatible signature.
    Two extrapolation methods are available:

    -   *Linear*: Linearly extrapolates given points using the slope defined by
        the interpolator boundaries (xi[0], xi[1]) if x < xi[0] and
        (xi[-1], xi[-2]) if x > xi[-1].
    -   *Constant*: Extrapolates given points by assigning the interpolator
        boundaries values xi[0] if x < xi[0] and xi[-1] if x > xi[-1].

    Specifying the *left* and *right* arguments takes precedence on the chosen
    extrapolation method and will assign the respective *left* and *right*
    values to the given points.

    Parameters
    ----------
    interpolator
        Interpolator object.
    method
        Extrapolation method.
    left
        Value to return for x < xi[0].
    right
        Value to return for x > xi[-1].
    dtype
        Data type used for internal conversions.

    Methods
    -------
    -   :meth:`~colour.Extrapolator.__init__`
    -   :meth:`~colour.Extrapolator.__class__`

    Notes
    -----
    -   The interpolator must define ``x`` and ``y`` properties.

    References
    ----------
    :cite:`Sastanina`, :cite:`Westland2012i`

    Examples
    --------
    Extrapolating a single numeric variable:

    >>> from colour.algebra import LinearInterpolator
    >>> x = np.array([3, 4, 5])
    >>> y = np.array([1, 2, 3])
    >>> interpolator = LinearInterpolator(x, y)
    >>> extrapolator = Extrapolator(interpolator)
    >>> extrapolator(1)
    -1.0

    Extrapolating an `ArrayLike` variable:

    >>> extrapolator(np.array([6, 7 , 8]))
    array([ 4.,  5.,  6.])

    Using the *Constant* extrapolation method:

    >>> x = np.array([3, 4, 5])
    >>> y = np.array([1, 2, 3])
    >>> interpolator = LinearInterpolator(x, y)
    >>> extrapolator = Extrapolator(interpolator, method='Constant')
    >>> extrapolator(np.array([0.1, 0.2, 8, 9]))
    array([ 1.,  1.,  3.,  3.])

    Using defined *left* boundary and *Constant* extrapolation method:

    >>> x = np.array([3, 4, 5])
    >>> y = np.array([1, 2, 3])
    >>> interpolator = LinearInterpolator(x, y)
    >>> extrapolator = Extrapolator(interpolator, method='Constant', left=0)
    >>> extrapolator(np.array([0.1, 0.2, 8, 9]))
    array([ 0.,  0.,  3.,  3.])
    """

    def __init__(
        self,
        interpolator: Optional[TypeInterpolator] = None,
        method: Union[Literal["Linear", "Constant"], str] = "Linear",
        left: Optional[Number] = None,
        right: Optional[Number] = None,
        dtype: Optional[Type[DTypeNumber]] = None,
    ):
        dtype = cast(Type[DTypeNumber], optional(dtype, DEFAULT_FLOAT_DTYPE))

        self._interpolator: TypeInterpolator = NullInterpolator(
            np.array([-np.inf, np.inf]), np.array([-np.inf, np.inf])
        )
        self.interpolator = optional(interpolator, self._interpolator)
        self._method: Union[Literal["Linear", "Constant"], str] = "Linear"
        self.method = cast(
            Union[Literal["Linear", "Constant"], str],
            optional(method, self._method),
        )
        self._right: Optional[Number] = None
        self.right = right
        self._left: Optional[Number] = None
        self.left = left

        self._dtype: Type[DTypeNumber] = dtype

    @property
    def interpolator(self) -> TypeInterpolator:
        """
        Getter and setter property for the *Colour* or *scipy* interpolator
        class instance.

        Parameters
        ----------
        value
            Value to set the *Colour* or *scipy* interpolator class instance
            with.

        Returns
        -------
        TypeInterpolator
            *Colour* or *scipy* interpolator class instance.
        """

        return self._interpolator

    @interpolator.setter
    def interpolator(self, value: TypeInterpolator):
        """Setter for the **self.interpolator** property."""

        attest(
            hasattr(value, "x"),
            f'"{value}" interpolator has no "x" attribute!',
        )

        attest(
            hasattr(value, "y"),
            f'"{value}" interpolator has no "y" attribute!',
        )

        self._interpolator = value

    @property
    def method(self) -> Union[Literal["Linear", "Constant"], str]:
        """
        Getter and setter property for the extrapolation method.

        Parameters
        ----------
        value
            Value to set the extrapolation method. with.

        Returns
        -------
        :class:`str`
            Extrapolation method.
        """

        return self._method

    @method.setter
    def method(self, value: Union[Literal["Linear", "Constant"], str]):
        """Setter for the **self.method** property."""

        attest(
            is_string(value),
            f'"method" property: "{value}" type is not "str"!',
        )

        value = validate_method(value, ["Linear", "Constant"])

        self._method = value

    @property
    def left(self) -> Optional[Number]:
        """
        Getter and setter property for left value to return for x < xi[0].

        Parameters
        ----------
        value
            Left value to return for x < xi[0].

        Returns
        -------
        :py:data:`None` or Number
            Left value to return for x < xi[0].
        """

        return self._left

    @left.setter
    def left(self, value: Optional[Number]):
        """Setter for the **self.left** property."""

        if value is not None:
            attest(
                is_numeric(value),
                f'"left" property: "{value}" is not a "number"!',
            )

            self._left = value

    @property
    def right(self) -> Optional[Number]:
        """
        Getter and setter property for right value to return for x > xi[-1].

        Parameters
        ----------
        value
            Right value to return for x > xi[-1].

        Returns
        -------
        :py:data:`None` or Number
            Right value to return for x > xi[-1].
        """

        return self._right

    @right.setter
    def right(self, value: Optional[Number]):
        """Setter for the **self.right** property."""

        if value is not None:
            attest(
                is_numeric(value),
                f'"right" property: "{value}" is not a "number"!',
            )

            self._right = value

    def __call__(self, x: FloatingOrArrayLike) -> FloatingOrNDArray:
        """
        Evaluate the Extrapolator at given point(s).

        Parameters
        ----------
        x
            Point(s) to evaluate the Extrapolator at.

        Returns
        -------
        :class:`numpy.floating` or :class:`numpy.ndarray`
            Extrapolated points value(s).
        """

        x = np.atleast_1d(x).astype(self._dtype)

        xe = as_float(self._evaluate(x))

        return xe

    def _evaluate(self, x: NDArray) -> NDArray:
        """
        Perform the extrapolating evaluation at given points.

        Parameters
        ----------
        x
            Points to evaluate the Extrapolator at.

        Returns
        -------
        :class:`numpy.ndarray`
            Extrapolated points values.
        """

        xi = self._interpolator.x
        yi = self._interpolator.y

        y = np.empty_like(x)

        if self._method == "linear":
            y[x < xi[0]] = yi[0] + (x[x < xi[0]] - xi[0]) * (yi[1] - yi[0]) / (
                xi[1] - xi[0]
            )
            y[x > xi[-1]] = yi[-1] + (x[x > xi[-1]] - xi[-1]) * (
                yi[-1] - yi[-2]
            ) / (xi[-1] - xi[-2])
        elif self._method == "constant":
            y[x < xi[0]] = yi[0]
            y[x > xi[-1]] = yi[-1]

        if self._left is not None:
            y[x < xi[0]] = self._left
        if self._right is not None:
            y[x > xi[-1]] = self._right

        in_range = np.logical_and(x >= xi[0], x <= xi[-1])
        y[in_range] = self._interpolator(x[in_range])

        return y
